from __future__ import annotations

import time
import typing
from typing import Any

import requests
from digitalhub_core.entities._base.crud import (
    list_entity_api_base,
    list_entity_api_ctx,
    logs_api,
    read_entity_api_ctx,
    stop_api,
)
from digitalhub_core.entities._base.entity.unversioned import UnversionedEntity
from digitalhub_core.entities._base.state import State
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.registry.registry import registry
from digitalhub_core.runtimes.builder import build_runtime
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.entity.material import MaterialEntity
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities.run.spec import RunSpec
    from digitalhub_core.entities.run.status import RunStatus
    from digitalhub_core.runtimes.base import Runtime


class Run(UnversionedEntity):
    """
    A class representing a run.
    """

    ENTITY_TYPE = EntityTypes.RUN.value

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: RunSpec,
        status: RunStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, uuid, kind, metadata, spec, status, user)

        self.spec: RunSpec
        self.status: RunStatus

    ##############################
    #  Run Methods
    ##############################

    def build(self) -> None:
        """
        Build run.

        Returns
        -------
        None
        """
        runtime = self._get_runtime()
        executable = self._get_executable(runtime)
        task = self._get_task(runtime)
        new_spec = runtime.build(executable, task, self.to_dict())
        self.spec = build_spec(
            self.kind,
            **new_spec,
        )
        self._set_state(State.BUILT.value)
        self.save()

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        self.refresh()
        if self.spec.local_execution:
            if not self._is_built():
                raise EntityError("Run is not in built state. Build it again.")
            self._set_state(State.RUNNING.value)
            self.save(update=True)

        # Try to get inputs if they exist
        try:
            self.spec.inputs = self.inputs(as_dict=True)
        except EntityError:
            pass

        try:
            status = self._get_runtime().run(self.to_dict())
        except Exception as e:
            self.refresh()
            if self.spec.local_execution:
                self._set_state(State.ERROR.value)
            self._set_message(str(e))
            self.save(update=True)
            raise e

        self.refresh()
        if not self.spec.local_execution:
            status.pop("state", None)
        new_status = {**self.status.to_dict(), **status}
        self._set_status(new_status)
        self.save(update=True)
        return self

    def wait(self, log_info: bool = True) -> Run:
        """
        Wait for run to finish.

        Parameters
        ----------
        log_info : bool
            If True, log information.

        Returns
        -------
        Run
            Run object.
        """
        start = time.time()
        while True:
            if log_info:
                LOGGER.info(f"Waiting for run {self.id} to finish...")
            self.refresh()
            time.sleep(5)
            if self.status.state in [
                State.STOPPED.value,
                State.ERROR.value,
                State.COMPLETED.value,
            ]:
                if log_info:
                    current = time.time() - start
                    LOGGER.info(f"Run {self.id} finished in {current:.2f} seconds.")
                return self

    def inputs(self, as_dict: bool = False) -> list[dict]:
        """
        Get inputs passed in spec as objects or as dictionaries.

        Parameters
        ----------
        as_dict : bool
            If True, return inputs as dictionaries.

        Returns
        -------
        list[dict]
            List of input objects.
        """
        try:
            return self.spec.get_inputs(as_dict=as_dict)
        except AttributeError:
            msg = f"Run of type {self.kind} has no inputs."
            raise EntityError(msg)

    def results(self) -> dict:
        """
        Get results from runtime execution.

        Returns
        -------
        dict
            Results.
        """
        try:
            return self.status.get_results()
        except AttributeError:
            msg = f"Run of type {self.kind} has no results."
            raise EntityError(msg)

    def result(self, key: str) -> Any:
        """
        Get result from runtime execution by key.

        Parameters
        ----------
        key : str
            Key of the result.

        Returns
        -------
        Any
            Result.
        """
        return self.results().get(key)

    def outputs(self, as_key: bool = False, as_dict: bool = False) -> dict:
        """
        Get run objects results.

        Parameters
        ----------
        as_key : bool
            If True, return results as keys.
        as_dict : bool
            If True, return results as dictionaries.

        Returns
        -------
        dict
            List of output objects.
        """
        try:
            return self.status.get_outputs(as_key=as_key, as_dict=as_dict)
        except AttributeError:
            msg = f"Run of type {self.kind} has no outputs."
            raise EntityError(msg)

    def output(self, key: str, as_key: bool = False, as_dict: bool = False) -> MaterialEntity | dict | str | None:
        """
        Get run object result by key.

        Parameters
        ----------
        key : str
            Key of the result.
        as_key : bool
            If True, return result as key.
        as_dict : bool
            If True, return result as dictionary.

        Returns
        -------
        Entity | dict | str | None
            Result.
        """
        return self.outputs(as_key=as_key, as_dict=as_dict).get(key)

    def values(self) -> dict:
        """
        Get values from runtime execution.

        Returns
        -------
        dict
            Values from backend.
        """
        try:
            value_list = getattr(self.spec, "values", [])
            value_list = value_list if value_list is not None else []
            return self.status.get_values(value_list)
        except AttributeError:
            msg = f"Run of type {self.kind} has no values."
            raise EntityError(msg)

    def logs(self) -> dict:
        """
        Get object from backend.
        Returns empty dictionary if context is local.

        Returns
        -------
        dict
            Logs from backend.
        """
        if self._context().local:
            return {}
        return logs_api(self.project, self.ENTITY_TYPE, self.id)

    def stop(self) -> None:
        """
        Stop run.

        Returns
        -------
        None
        """
        if not self._context().local and not self.spec.local_execution:
            return stop_api(self.project, self.ENTITY_TYPE, self.id)
        try:
            self.status.stop()
        except AttributeError:
            return

    def invoke(self, **kwargs) -> requests.Response:
        """
        Invoke run.

        Parameters
        ----------
        kwargs
            Keyword arguments to pass to the request.

        Returns
        -------
        requests.Response
            Response from service.
        """
        try:
            if not self._context().local and not self.spec.local_execution:
                local = False
            else:
                local = True
            if kwargs is None:
                kwargs = {}
            return self.status.invoke(local, **kwargs)
        except AttributeError:
            msg = f"Run of type {self.kind} has no invoke operation."
            raise EntityError(msg)

    ##############################
    #  Helpers
    ##############################

    def _is_built(self) -> bool:
        """
        Check if run is in built state.

        Returns
        -------
        bool
            True if run is in built state, False otherwise.
        """
        return self.status.state == State.BUILT.value

    def _set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.

        Returns
        -------
        None
        """
        self.status: RunStatus = build_status(self.kind, **status)

    def _set_state(self, state: str) -> None:
        """
        Update run state.

        Parameters
        ----------
        state : str
            State to set.

        Returns
        -------
        None
        """
        self.status.state = state

    def _set_message(self, message: str) -> None:
        """
        Update run message.

        Parameters
        ----------
        message : str
            Message to set.

        Returns
        -------
        None
        """
        self.status.message = message

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        return build_runtime(self.kind, self.project)

    def _get_executable(self, runtime: Runtime) -> dict:
        """
        Get object from backend. Reimplemented to avoid circular imports.

        Parameters
        ----------
        runtime : Runtime
            Runtime object.

        Returns
        -------
        dict
            Executable (function or workflow) from backend.
        """
        exec_kind = runtime.get_executable_kind()
        entity_type = registry.get_entity_type(exec_kind)
        splitted = self.spec.task.split("/")
        exec_name = splitted[-1].split(":")[0]
        exec_id = splitted[-1].split(":")[1]
        return read_entity_api_ctx(
            exec_name,
            entity_type=entity_type,
            project=self.project,
            entity_id=exec_id,
        )

    def _get_task(self, runtime: Runtime) -> dict:
        """
        Get object from backend. Reimplemented to avoid circular imports.

        Parameters
        ----------
        runtime : Runtime
            Runtime object.

        Returns
        -------
        dict
            Task from backend.
        """
        executable_kind = runtime.get_executable_kind()
        exec_string = f"{executable_kind}://{self.spec.task.split('://')[1]}"

        # Local backend
        if self._context().local:
            tasks = list_entity_api_base(self._context().client, EntityTypes.TASK.value)
            for i in tasks:
                if i.get("spec").get("function") == exec_string:
                    return i
            raise EntityError("Task not found.")

        # Remote backend
        task_kind = self.spec.task.split("://")[0]
        params = {"function": exec_string, "kind": task_kind}
        return list_entity_api_ctx(self.project, EntityTypes.TASK.value, params=params)[0]



from __future__ import annotations

import typing

from digitalhub_core.entities._base.state import State
from digitalhub_core.registry.registry import registry
from digitalhub_core.registry.utils import import_class

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.status.base import Status
    from digitalhub_core.registry.models import RegistryEntry


def build_status(kind: str, **kwargs) -> Status:
    """
    Build entity status object. The builder takes as input
    the kind of status's object to build and the keyword
    arguments to pass to the status's constructor.
    The specific Status class is searched in the global
    registry, where lies info about where to find the class.
    The arguments are parsed, eventually adding default values,
    and then passed to the constructor.

    Parameters
    ----------
    kind : str
        Registry entry kind.
    **kwargs : dict
        Keyword arguments for the constructor.

    Returns
    -------
    Status
        Status object.
    """
    infos: RegistryEntry = getattr(registry, kind)
    status = import_class(infos.status.module, infos.status.class_name)
    kwargs = parse_arguments(**kwargs)
    return status(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values if necessary.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    dict
        Keyword arguments with default values.
    """
    state = kwargs.get("state")
    if state is None:
        kwargs["state"] = State.CREATED.value
    else:
        if kwargs["state"] not in State.__members__:
            raise ValueError(f"Invalid state: {state}")
    return kwargs



from __future__ import annotations

from pydantic import BaseModel


class RuntimeEntry(BaseModel):
    """
    Model for Runtime entry.
    It specifies the module path and the runtime class name.
    """

    module: str
    """Module path."""

    class_name: str
    """Class name."""

    kind_registry_module: str
    """Kind registry module path."""

    kind_registry_class_name: str
    """Kind registry class name."""


class SpecEntry(BaseModel):
    """
    Model for Spec entry.
    It specifies the module path, the spec class name
    and the spec pydantic class validator.
    """

    module: str
    """Module path."""

    class_name: str
    """Class name."""

    parameters_validator: str
    """Class name of the parameter validator."""


class StatusEntry(BaseModel):
    """
    Model for Status entry.
    It specifies the module path and the status class name.
    """

    module: str
    """Module path."""

    class_name: str
    """Class name."""


class MetadataEntry(BaseModel):
    """
    Model for Metadata entry.
    It specifies the module path and the metadata class name.
    """

    module: str
    """Module path."""

    class_name: str
    """Class name."""


class RegistryEntry(BaseModel):
    """
    Basic entry model for every entity.
    Functions, tasks and runs specify also the runtime field.
    """

    entity_type: str
    """Entity type."""

    spec: SpecEntry
    """Spec infos."""

    status: StatusEntry
    """Status infos."""

    metadata: MetadataEntry
    """Metadata infos."""

    runtime: RuntimeEntry = None
    """Runtime infos."""



from __future__ import annotations

import importlib.util as imputil
import typing
from pathlib import Path

from digitalhub_core.client.builder import build_client, get_client
from digitalhub_core.context.builder import delete_context
from digitalhub_core.entities._base.crud import delete_entity_api_base, read_entity_api_base, update_entity_api_base
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.entities.project.builder import project_from_dict, project_from_parameters
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.project.entity._base import Project


ENTITY_TYPE = EntityTypes.PROJECT.value


def new_project(
    name: str,
    description: str | None = None,
    labels: list[str] | None = None,
    local: bool = False,
    config: dict | None = None,
    context: str | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Create a new object.

    Parameters
    ----------
    name : str
        Object name.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    local : bool
        If True, use local backend, if False use DHCore backend. Default to False.
    config : dict
        DHCore environment configuration.
    context : str
        The context local folder of the project.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = new_project("my-project")
    """
    build_client(local, config)
    if context is None:
        context = name
    obj = project_from_parameters(
        name=name,
        kind="project",
        description=description,
        labels=labels,
        local=local,
        context=context,
        **kwargs,
    )
    obj.save()
    return _setup_project(obj, setup_kwargs)


def get_project(
    name: str,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Retrieves project details from backend.

    Parameters
    ----------
    name : str
        The Project name.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore environment configuration.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = get_project("my-project")
    """
    build_client(local, config)
    client = get_client(local)
    obj = read_entity_api_base(client, ENTITY_TYPE, name, **kwargs)
    obj["local"] = local
    project = project_from_dict(obj)
    return _setup_project(project, setup_kwargs)


def import_project(
    file: str,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Import object from a YAML file.

    Parameters
    ----------
    file : str
        Path to YAML file.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore environment configuration.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = import_project("my-project.yaml")
    """
    build_client(local, config)
    obj: dict = read_yaml(file)
    obj["local"] = local
    project = project_from_dict(obj)
    return _setup_project(project, setup_kwargs)


def load_project(
    name: str | None = None,
    filename: str | None = None,
    local: bool = False,
    config: dict | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Load project and context from backend or file. Name or
    filename must be provided. Name takes precedence over filename.

    Parameters
    ----------
    name : str
        Project name.
    filename : str
        Path to YAML file.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore environment configuration.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    If name is provided, load project from backend.
    >>> obj = load_project(name="my-project")

    If filename is provided, load project from file.
    >>> obj = load_project(filename="my-project.yaml")
    """
    if name is not None:
        return get_project(name=name, local=local, config=config, setup_kwargs=setup_kwargs, **kwargs)
    if filename is not None:
        return import_project(filename, local=local, config=config, setup_kwargs=setup_kwargs)
    raise EntityError("Either name or filename must be provided.")


def get_or_create_project(
    name: str,
    local: bool = False,
    config: dict | None = None,
    context: str | None = None,
    setup_kwargs: dict | None = None,
    **kwargs,
) -> Project:
    """
    Try to get project. If not exists, create it.

    Parameters
    ----------
    name : str
        Project name.
    local : bool
        Flag to determine if backend is local.
    config : dict
        DHCore environment configuration.
    context : str
        Folder where the project will saves its context locally.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    Project
        Object instance.
    """
    try:
        return get_project(
            name,
            local=local,
            config=config,
            setup_kwargs=setup_kwargs,
            **kwargs,
        )
    except BackendError:
        return new_project(
            name,
            local=local,
            config=config,
            setup_kwargs=setup_kwargs,
            context=context,
            **kwargs,
        )


def update_project(entity: Project, local: bool = False, **kwargs) -> Project:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Project
        Object to update.
    local : bool
        Flag to determine if backend is local.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Project
        The updated object.

    Examples
    --------
    >>> obj = update_project(obj)
    """
    client = get_client(local)
    obj = update_entity_api_base(client, ENTITY_TYPE, entity.name, entity.to_dict(), **kwargs)
    return project_from_dict(obj)


def delete_project(
    name: str,
    cascade: bool = True,
    clean_context: bool = True,
    local: bool = False,
    **kwargs,
) -> list[dict]:
    """
    Delete a project.

    Parameters
    ----------
    name : str
        Project name.
    cascade : bool
        Flag to determine if delete is cascading.
    clean_context : bool
        Flag to determine if context will be deleted. If a context is deleted,
        all its objects are unreacheable.
    local : bool
        Flag to determine if backend is local.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    >>> delete_project("my-project")
    """
    client = get_client(local)
    obj = delete_entity_api_base(client, ENTITY_TYPE, name, cascade=cascade, **kwargs)
    if clean_context:
        delete_context(name)
    return obj


def _setup_project(project: Project, setup_kwargs: dict | None = None) -> Project:
    """
    Search for setup_project.py file and launch setup hanlder as project hook.

    Parameters
    ----------
    project : Project
        The project to scafold.
    setup_kwargs : dict
        Arguments to pass to setup handler.

    Returns
    -------
    Project
        Set up project.
    """
    setup_kwargs = setup_kwargs if setup_kwargs is not None else {}
    check_pth = Path(project.spec.context, ".CHECK")
    setup_pth = Path(project.spec.context, "setup_project.py")
    if setup_pth.exists() and not check_pth.exists():
        spec = imputil.spec_from_file_location("setup_project", setup_pth)
        mod = imputil.module_from_spec(spec)
        spec.loader.exec_module(mod)
        handler = getattr(mod, "setup")
        project = handler(project, **setup_kwargs)
        check_pth.touch()
    return project



from __future__ import annotations

from enum import Enum


class EntityTypes(Enum):
    """
    Entity types.
    """

    PROJECT = "project"
    ARTIFACT = "artifact"
    SECRET = "secret"
    FUNCTION = "function"
    WORKFLOW = "workflow"
    TASK = "task"
    RUN = "run"



from __future__ import annotations

import typing

from digitalhub_core.registry.registry import registry
from digitalhub_core.registry.utils import import_class

if typing.TYPE_CHECKING:
    from digitalhub_core.registry.models import RegistryEntry
    from digitalhub_core.runtimes.base import Runtime
    from digitalhub_core.runtimes.kind_registry import KindRegistry


def build_runtime(kind: str, project: str) -> Runtime:
    """
    Build runtime object. The builder takes in input a kind.
    This kind can derive from functions, tasks, or runs, and
    is inserted in the global kind registry where the runtimes
    pakages are registered.
    The builder requires the module path where the Runtime
    subclass is defined and the class name. It requires also
    the kind registry module, kind registry class and the project
    name.

    Parameters
    ----------
    kind : str
        The type of runtime to build.
    project : str
        The project name.

    Returns
    -------
    Runtime
        Runtime object.
    """
    infos: RegistryEntry = getattr(registry, kind)
    cls = import_class(infos.runtime.module, infos.runtime.class_name)
    kind_registry = get_kind_registry(kind)
    return cls(kind_registry, project)


def get_kind_registry(kind: str) -> KindRegistry:
    """
    Get kind registry.

    Returns
    -------
    KindRegistry
        Kind registry.
    """
    infos: RegistryEntry = getattr(registry, kind)
    return import_class(infos.runtime.kind_registry_module, infos.runtime.kind_registry_class_name)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from tempfile import mkdtemp
from typing import Literal

from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.uri_utils import map_uri_scheme
from pydantic import BaseModel


class Store(metaclass=ABCMeta):
    """
    Store abstract class.
    """

    def __init__(self, name: str, store_type: str) -> None:
        """
        Constructor.

        Parameters
        ----------
        name : str
            Store name.
        store_type : str
            Store type. Used to choose the right store implementation.

        Returns
        -------
        None
        """
        self.name = name
        self.type = store_type

    ##############################
    # IO methods
    ##############################

    @abstractmethod
    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Method to download artifact from storage.
        """

    @abstractmethod
    def upload(self, src: str | list[str], dst: str | None = None) -> list[tuple[str, str]]:
        """
        Method to upload artifact to storage.
        """

    @abstractmethod
    def get_file_info(self, paths: list[str]) -> list[dict]:
        """
        Method to get file metadata.
        """

    ##############################
    # Helpers methods
    ##############################

    def _check_local_src(self, src: str) -> None:
        """
        Check if the source path is local.

        Parameters
        ----------
        src : str
            The source path.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If the source is not a local path.
        """
        if map_uri_scheme(src) != "local":
            raise StoreError(f"Source '{src}' is not a local path.")

    def _check_local_dst(self, dst: str) -> None:
        """
        Check if the destination path is local.

        Parameters
        ----------
        dst : str
            The destination path.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If the destination is not a local path.
        """
        if map_uri_scheme(dst) != "local":
            raise StoreError(f"Destination '{dst}' is not a local path.")

    def _check_overwrite(self, dst: Path, overwrite: bool) -> None:
        """
        Check if destination path exists for overwrite.

        Parameters
        ----------
        dst : Path
            Destination path as filename.
        overwrite : bool
            Specify if overwrite an existing file.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If destination path exists and overwrite is False.
        """
        if dst.exists() and not overwrite:
            raise StoreError(f"Destination {str(dst)} already exists.")

    @staticmethod
    def _build_path(path: str | Path) -> None:
        """
        Get path from store path and path.

        Parameters
        ----------
        path : str
            The path to build.

        Returns
        -------
        None
        """
        if not isinstance(path, Path):
            path = Path(path)
        if path.suffix != "":
            path = path.parent
        path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _build_temp() -> Path:
        """
        Build a temporary path.

        Returns
        -------
        Path
            Temporary path.
        """
        tmpdir = mkdtemp()
        return Path(tmpdir)


class StoreConfig(BaseModel):
    """
    Store configuration base class.
    """


class StoreParameters(BaseModel):
    """
    Store configuration class.
    """

    name: str
    """Store id."""

    type: Literal["local", "s3", "remote", "sql"]
    """Store type to instantiate."""

    config: StoreConfig = None
    """Configuration for the store."""

    is_default: bool = False
    """Flag to determine if the store is the default one."""



from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.crud import create_entity_api_ctx, read_entity_api_ctx, update_entity_api_ctx
from digitalhub_core.entities._base.entity.base import Entity
from digitalhub_core.utils.generic_utils import get_timestamp

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities._base.spec.base import Spec
    from digitalhub_core.entities._base.status.base import Status


class ContextEntity(Entity):
    def __init__(
        self,
        project: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        super().__init__(kind, metadata, spec, status, user)
        self.project = project
        self._obj_attr.extend(["project"])

    ##############################
    #  Save / Refresh / Export
    ##############################

    def save(self, update: bool = False) -> ContextEntity:
        """
        Save or update the entity into the backend.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        ContextEntity
            Entity saved.
        """
        obj = self.to_dict()
        if not update:
            return self._save(obj)
        return self._update(obj)

    def _save(self, obj: dict) -> ContextEntity:
        """
        Save entity into backend.

        Parameters
        ----------
        obj : dict
            Object instance as dictionary.

        Returns
        -------
        ContextEntity
            Entity saved.
        """
        new_obj = create_entity_api_ctx(self.project, self.ENTITY_TYPE, obj)
        self._update_attributes(new_obj)
        return self

    def _update(self, obj: dict) -> ContextEntity:
        """
        Update entity in backend.

        Parameters
        ----------
        obj : dict
            Object instance as dictionary.

        Returns
        -------
        ContextEntity
            Entity updated.
        """
        if self._context().local:
            self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        new_obj = update_entity_api_ctx(self.project, self.ENTITY_TYPE, self.id, obj)
        self._update_attributes(new_obj)
        return self

    def refresh(self) -> ContextEntity:
        """
        Refresh object from backend.

        Returns
        -------
        ContextEntity
            Entity refreshed.
        """
        new_obj = read_entity_api_ctx(self.key)
        self._update_attributes(new_obj)
        return self

    ##############################
    #  Context
    ##############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context object.
        """
        return get_context(self.project)



from __future__ import annotations


class StoreError(Exception):
    """
    Raised when incontered errors on stores.
    """


class BackendError(Exception):
    """
    Raised when incontered errors from backend.
    """


class EntityError(Exception):
    """
    Raised when incontered errors on entities.
    """



from __future__ import annotations

from pathlib import Path

import yaml

##############################
# Writers
##############################


def write_yaml(filepath: str | Path, obj: dict | list[dict]) -> None:
    """
    Write a dict or a list of dict to a yaml file.

    Parameters
    ----------
    filepath : str | Path
        The yaml file path to write.
    obj : dict
        The dict to write.

    Returns
    -------
    None
    """
    if isinstance(obj, list):
        with open(filepath, "w", encoding="utf-8") as out_file:
            yaml.dump_all(obj, out_file, sort_keys=False, default_flow_style=False)
    else:
        with open(filepath, "w", encoding="utf-8") as out_file:
            yaml.dump(obj, out_file, sort_keys=False)


##############################
# Readers
##############################


def read_yaml(filepath: str | Path) -> dict | list[dict]:
    """
    Read a yaml file and return a dict or a list of dict.

    Parameters
    ----------
    filepath : str | Path
        The yaml file path to read.

    Returns
    -------
    dict | list[dict]
        The yaml file content.
    """
    try:
        with open(filepath, "r", encoding="utf-8") as in_file:
            data = yaml.load(in_file, Loader=yaml.SafeLoader)

    # If yaml contains multiple documents
    except yaml.composer.ComposerError:
        with open(filepath, "r", encoding="utf-8") as in_file:
            data = list(yaml.load_all(in_file, Loader=yaml.SafeLoader))
    return data


def read_text(filepath: str | Path) -> str:
    """
    Read a file and return the text.

    Parameters
    ----------
    filepath : str | Path
        The file path to read.

    Returns
    -------
    str
        The file content.
    """
    return Path(filepath).read_text(encoding="utf-8")



from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj


class Metadata(ModelObj):
    """
    A class representing the metadata of an entity.
    Metadata is a collection of information about an entity thought
    to be modifiable by the user. The information contained in the
    metadata can be discordant with the actual state of the entity,
    for example the name of the entity in the database.
    """

    def __init__(
        self,
        project: str | None = None,
        name: str | None = None,
        version: str | None = None,
        description: str | None = None,
        labels: list[str] | None = None,
        created: str | None = None,
        created_by: str | None = None,
        updated: str | None = None,
        updated_by: str | None = None,
        embedded: bool | None = None,
        **kwargs,
    ) -> None:
        self.project = project
        self.name = name
        self.version = version
        self.description = description
        self.labels = labels
        self.created = created
        self.updated = updated
        self.created_by = created_by
        self.updated_by = updated_by
        self.embedded = embedded

        self._any_setter(**kwargs)

    @classmethod
    def from_dict(cls, obj: dict) -> Metadata:
        """
        Return entity metadata object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity metadata.

        Returns
        -------
        Metadata
            An entity metadata object.
        """
        return cls(**obj)



from __future__ import annotations

import typing

from digitalhub_core.context.builder import check_context
from digitalhub_core.entities._base.crud import (
    delete_entity_api_ctx,
    list_entity_api_ctx,
    read_entity_api_ctx,
    read_entity_api_ctx_versions,
)
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.entities.workflow.builder import workflow_from_dict, workflow_from_parameters
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.workflow.entity import Workflow


ENTITY_TYPE = EntityTypes.WORKFLOW.value


def new_workflow(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = True,
    **kwargs,
) -> Workflow:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    uuid : str
        ID of the object (UUID4, e.g. 40f25c4b-d26b-4221-b048-9527aff291e2).
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    >>> obj = new_function(project="my-project",
    >>>                    name="my-function",
    >>>                    kind="kfp",
    >>>                    code_src="pipeline.py",
    >>>                    handler="pipeline-handler")
    """
    check_context(project)
    obj = workflow_from_parameters(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        **kwargs,
    )
    obj.save()
    return obj


def get_workflow(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> Workflow:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_workflow("store://my-workflow-key")

    Using entity name:
    >>> obj = get_workflow("my-workflow-name"
    >>>                    project="my-project",
    >>>                    entity_id="my-workflow-id")
    """
    obj = read_entity_api_ctx(
        identifier,
        ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )
    return workflow_from_dict(obj)


def get_workflow_versions(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> list[Workflow]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Workflow]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> obj = get_workflow_versions("store://my-workflow-key")

    Using entity name:
    >>> obj = get_workflow_versions("my-workflow-name"
    >>>                             project="my-project")
    """
    obj = read_entity_api_ctx_versions(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        **kwargs,
    )
    return [workflow_from_dict(o) for o in obj]


def list_workflows(project: str, **kwargs) -> list[Workflow]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Workflow]
        List of object instances.

    Examples
    --------
    >>> objs = list_workflows(project="my-project")
    """
    objs = list_entity_api_ctx(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )
    return [workflow_from_dict(obj) for obj in objs]


def import_workflow(file: str) -> Workflow:
    """
    Import object from a YAML file.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    >>> obj = import_workflow("my-workflow.yaml")
    """
    obj: dict = read_yaml(file)
    if isinstance(obj, list):
        wf_dict = obj[0]
        task_dicts = obj[1:]
    else:
        wf_dict = obj
        task_dicts = []

    check_context(obj.get("project"))
    workflow = workflow_from_dict(wf_dict)
    workflow.import_tasks(task_dicts)
    return workflow


def update_workflow(entity: Workflow) -> Workflow:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Workflow
        Object to update.

    Returns
    -------
    Workflow
        Entity updated.

    Examples
    --------
    >>> obj = update_workflow(obj)
    """
    return entity.save(update=True)


def delete_workflow(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_workflow("store://my-workflow-key")

    Otherwise:
    >>> obj = delete_workflow("workflow-name",
    >>>                       project="my-project",
    >>>                       delete_all_versions=True)
    """
    return delete_entity_api_ctx(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
        **kwargs,
    )



from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod

from digitalhub_core.entities._base.base import ModelObj

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities._base.spec.base import Spec
    from digitalhub_core.entities._base.status.base import Status


class Entity(ModelObj, metaclass=ABCMeta):
    """
    Abstract class for entities.

    An entity is a collection of metadata, specification and status
    representing a variety of objects handled by Digitalhub.
    """

    # Entity type
    # Need to be set in subclasses
    ENTITY_TYPE: str

    # Attributes to render as dict. Need to be expanded in subclasses.
    _obj_attr = ["kind", "metadata", "spec", "status", "user", "key"]

    def __init__(
        self,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status
        self.user = user

        # Need to be set in subclasses
        self.key: str

    @abstractmethod
    def save(self, update: bool = False) -> Entity:
        """
        Abstract save method.
        """

    @abstractmethod
    def refresh(self) -> Entity:
        """
        Abstract refresh method.
        """

    def _update_attributes(self, obj: dict) -> None:
        """
        Update attributes.

        Parameters
        ----------
        obj : dict
            Mapping representation of object.

        Returns
        -------
        None
        """
        new_obj = self.from_dict(obj)
        self.metadata = new_obj.metadata
        self.spec = new_obj.spec
        self.status = new_obj.status
        self.user = new_obj.user

    @abstractmethod
    def export(self, filename: str | None = None) -> None:
        """
        Abstract export method.
        """

    def to_dict(self) -> dict:
        """
        Override default to_dict method to add the possibility to exclude
        some attributes. This requires to set a list of _obj_attr
        attributes in the subclass.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        return {k: v for k, v in super().to_dict().items() if k in self._obj_attr}

    @classmethod
    def from_dict(cls, obj: dict, validate: bool = True) -> Entity:
        """
        Create a new object from dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary to create object from.
        validate : bool
            Flag to indicate if arguments validation against a pydantic schema must be ignored.

        Returns
        -------
        Self
            Self instance.
        """
        parsed_dict = cls._parse_dict(obj, validate=validate)
        return cls(**parsed_dict)

    @staticmethod
    @abstractmethod
    def _parse_dict(obj: dict, validate: bool = True) -> dict:
        """
        Abstract method to parse dictionary.
        """

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict())



from __future__ import annotations

API_BASE = "/api/v1"
API_CONTEXT = f"{API_BASE}/-"


##############################
# Context APIs
##############################


def api_ctx_create(
    project: str,
    entity_type: str,
) -> str:
    """
    Create context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        The name of the entity_type.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}"


def api_ctx_read(project: str, entity_type: str, entity_id: str) -> str:
    """
    Read context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}"


def api_ctx_update(
    project: str,
    entity_type: str,
    entity_id: str,
) -> str:
    """
    Update context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}"


def api_ctx_delete(project: str, entity_type: str, entity_id: str) -> str:
    """
    Delete context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}"


def api_ctx_list(project: str, entity_type: str) -> str:
    """
    List context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}"


def api_ctx_data(project: str, entity_type: str) -> str:
    """
    Data context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/data"


def api_ctx_logs(project: str, entity_type: str, entity_id: str) -> str:
    """
    Logs context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}/logs"


def api_ctx_stop(project: str, entity_type: str, entity_id: str) -> str:
    """
    Stop context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}/stop"


def api_ctx_files(project: str, entity_type: str, entity_id: str) -> str:
    """
    Files context API.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_CONTEXT}/{project}/{entity_type}/{entity_id}/files/info"


##############################
# Base APIs
##############################


def api_base_list(entity_type: str) -> str:
    """
    List base API.

    Parameters
    ----------
    entity_type : str
        Entity type.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_BASE}/{entity_type}"


def api_base_create(entity_type: str) -> str:
    """
    Create base API.

    Parameters
    ----------
    entity_type : str
        Entity type.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_BASE}/{entity_type}"


def api_base_read(entity_type: str, entity_id: str) -> str:
    """
    Read base API.

    Parameters
    ----------
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_BASE}/{entity_type}/{entity_id}"


def api_base_update(entity_type: str, entity_id: str) -> str:
    """
    Update base API.

    Parameters
    ----------
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_BASE}/{entity_type}/{entity_id}"


def api_base_delete(entity_type: str, entity_id: str) -> str:
    """
    Delete base API.

    Parameters
    ----------
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    str
        The API string formatted.
    """
    if not entity_type.endswith("s"):
        entity_type += "s"
    return f"{API_BASE}/{entity_type}/{entity_id}"



from __future__ import annotations

import typing
from pathlib import Path

from digitalhub_core.context.builder import check_context
from digitalhub_core.entities._base.crud import (
    delete_entity_api_ctx,
    list_entity_api_ctx,
    read_entity_api_ctx,
    read_entity_api_ctx_versions,
)
from digitalhub_core.entities._builders.uuid import build_uuid
from digitalhub_core.entities.artifact.builder import artifact_from_dict, artifact_from_parameters
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.utils.env_utils import get_s3_bucket
from digitalhub_core.utils.io_utils import read_yaml
from digitalhub_core.utils.uri_utils import check_local_path

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.artifact.entity._base import Artifact


ENTITY_TYPE = EntityTypes.ARTIFACT.value


def new_artifact(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = True,
    path: str | None = None,
    **kwargs,
) -> Artifact:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object (UUID4, e.g. 40f25c4b-d26b-4221-b048-9527aff291e2).
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    path : str
        Object path on local file system or remote storage. It is also the destination path of upload() method.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Artifact
        Object instance.

    Examples
    --------
    >>> obj = new_artifact(project="my-project",
    >>>                    name="my-artifact",
    >>>                    kind="artifact",
    >>>                    path="s3://my-bucket/my-key")
    """
    check_context(project)
    obj = artifact_from_parameters(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        path=path,
        **kwargs,
    )
    obj.save()
    return obj


def log_artifact(
    project: str,
    name: str,
    kind: str,
    source: list[str] | str,
    path: str | None = None,
    **kwargs,
) -> Artifact:
    """
    Create and upload an object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    source : str
        Artifact location on local path.
    path : str
        Destination path of the artifact. If not provided, it's generated.
    **kwargs : dict
        New artifact spec parameters.

    Returns
    -------
    Artifact
        Object instance.

    Examples
    --------
    >>> obj = log_artifact(project="my-project",
    >>>                    name="my-artifact",
    >>>                    kind="artifact",
    >>>                    source="./local-path")
    """
    if isinstance(source, list):
        source_is_local = all(check_local_path(s) for s in source)
        for s in source:
            if Path(s).is_dir():
                raise ValueError(f"Invalid source path: {s}. List of paths must be list of files, not directories.")
    else:
        source_is_local = check_local_path(source)

        if not source_is_local:
            raise ValueError(f"Invalid source path: {s}")

    if path is not None:
        if not source_is_local:
            raise ValueError("If you provide a path, you must use a local path as source.")
    else:
        if source_is_local:
            uuid = build_uuid()
            kwargs["uuid"] = uuid
            path = f"s3://{get_s3_bucket()}/{project}/{ENTITY_TYPE}/{name}/{uuid}"

            if isinstance(source, list) or Path(source).is_dir():
                path = f"{path}/"
            elif Path(source).is_file():
                path = f"{path}/{Path(source).name}"
            else:
                raise ValueError(f"Invalid source path: {source}")

        else:
            path = source

    obj = new_artifact(project=project, name=name, kind=kind, path=path, **kwargs)

    if source_is_local:
        obj.upload(source)
    return obj


def get_artifact(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> Artifact:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Artifact
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_artifact("store://my-artifact-key")

    Using entity name:
    >>> obj = get_artifact("my-artifact-name"
    >>>                    project="my-project",
    >>>                    entity_id="my-artifact-id")
    """
    obj = read_entity_api_ctx(
        identifier,
        ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )
    entity = artifact_from_dict(obj)
    entity._get_files_info()
    return entity


def get_artifact_versions(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> list[Artifact]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Artifact]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> obj = get_artifact_versions("store://my-artifact-key")

    Using entity name:
    >>> obj = get_artifact_versions("my-artifact-name"
    >>>                             project="my-project")
    """
    objs = read_entity_api_ctx_versions(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        **kwargs,
    )
    objects = []
    for o in objs:
        entity = artifact_from_dict(o)
        entity._get_files_info()
        objects.append(entity)
    return objects


def list_artifacts(project: str, **kwargs) -> list[Artifact]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Artifact]
        List of object instances.

    Examples
    --------
    >>> objs = list_artifacts(project="my-project")
    """
    objs = list_entity_api_ctx(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )
    objects = []
    for o in objs:
        entity = artifact_from_dict(o)
        entity._get_files_info()
        objects.append(entity)
    return objects


def import_artifact(file: str) -> Artifact:
    """
    Import object from a YAML file.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Artifact
        Object instance.

    Examples
    --------
    >>> obj = import_artifact("my-artifact.yaml")
    """
    obj: dict = read_yaml(file)
    return artifact_from_dict(obj)


def update_artifact(entity: Artifact) -> Artifact:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Artifact
        Object to update.

    Returns
    -------
    Artifact
        Entity updated.

    Examples
    --------
    >>> obj = update_artifact(obj)
    """
    return entity.save(update=True)


def delete_artifact(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> delete_artifact("store://my-artifact-key")

    Otherwise:
    >>> delete_artifact("my-artifact-name",
    >>>                  project="my-project",
    >>>                  delete_all_versions=True)
    """
    return delete_entity_api_ctx(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        **kwargs,
    )



from __future__ import annotations

from digitalhub_core.entities._base.spec.base import Spec, SpecParams


class FunctionSpec(Spec):
    """
    Specification for a Function.
    """


class FunctionParams(SpecParams):
    """
    Function parameters model.
    """


class SourceCodeStruct:
    """
    Source code struct.
    """

    def __init__(
        self,
        source: str | None = None,
        handler: str | None = None,
        code: str | None = None,
        base64: str | None = None,
        init_function: str | None = None,
        lang: str | None = None,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        source : str
            Source reference.
        handler : str
            Function entrypoint.
        code : str
            Source code (plain).
        base64 : str
            Source code (base64 encoded).
        init_function : str
            Init function for remote execution.
        lang : str
            Source code language (hint).
        """
        self.source = source
        self.handler = handler
        self.code = code
        self.base64 = base64
        self.init_function = init_function
        self.lang = lang

    def to_dict(self) -> dict:
        """
        Convert to dictionary.

        Returns
        -------
        dict
            Dictionary representation of the object.
        """
        dict_ = {}
        if self.source is not None:
            dict_["source"] = self.source
        if self.handler is not None:
            dict_["handler"] = self.handler
        if self.base64 is not None:
            dict_["base64"] = self.base64
        if self.init_function is not None:
            dict_["init_function"] = self.init_function
        if self.lang is not None:
            dict_["lang"] = self.lang

        return dict_

    def __repr__(self) -> str:
        return str(self.to_dict())



from __future__ import annotations

import typing

from digitalhub_core.entities._base.status.base import Status
from digitalhub_core.entities.artifact.crud import get_artifact
from digitalhub_core.entities.utils import parse_entity_key

if typing.TYPE_CHECKING:
    pass

ENTITY_FUNC = {
    "artifacts": get_artifact,
}


class RunStatus(Status):
    """
    Status class for run entities.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        outputs: list | None = None,
        results: dict | None = None,
        **kwargs,
    ) -> None:
        super().__init__(state, message)
        self.outputs = outputs
        self.results = results

        self._any_setter(**kwargs)

    def get_results(self) -> dict:
        """
        Get results.

        Returns
        -------
        dict
            The results.
        """
        if not hasattr(self, "results") or self.results is None:
            return {}
        return self.results

    def get_outputs(self, as_key: bool = False, as_dict: bool = False) -> dict:
        """
        Get outputs.

        Parameters
        ----------
        as_key : bool
            If True, return outputs as keys.
        as_dict : bool
            If True, return outputs as dictionaries.

        Returns
        -------
        dict
            The outputs.
        """
        outputs = {}
        if not hasattr(self, "outputs") or self.outputs is None:
            return outputs

        for parameter, key in self.outputs.items():
            entity_type = self._get_entity_type(key)
            entity = ENTITY_FUNC[entity_type](key)
            if as_key:
                entity = entity.key
            if as_dict:
                entity = entity.to_dict()
            outputs[parameter] = entity

        return outputs

    @staticmethod
    def _get_entity_type(key: str) -> str:
        """
        Get entity type.

        Parameters
        ----------
        key : str
            The key of the entity.

        Returns
        -------
        str
            The entity type.
        """
        _, entity_type, _, _, _ = parse_entity_key(key)
        return entity_type

    def get_values(self, values_list: list) -> dict:
        """
        Get values.

        Parameters
        ----------
        values_list : list
            The values list to search in.

        Returns
        -------
        dict
            The values.
        """
        if not hasattr(self, "results") or self.results is None:
            return {}
        return {k: v for k, v in self.get_results().items() if k in values_list}



from __future__ import annotations

import typing

from digitalhub_core.entities._base.entity.context import ContextEntity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.name import build_name
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities._builders.uuid import build_uuid
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities._base.spec.base import Spec
    from digitalhub_core.entities._base.status.base import Status


class VersionedEntity(ContextEntity):
    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        super().__init__(project, kind, metadata, spec, status, user)
        self.name = name
        self.id = uuid
        self.key = f"store://{project}/{self.ENTITY_TYPE}/{kind}/{name}:{uuid}"

        # Add attributes to be used in the to_dict method
        self._obj_attr.extend(["name", "id"])

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        if filename is None:
            filename = f"{self.ENTITY_TYPE}-{self.name}-{self.id}.yml"
        pth = self._context().root / filename
        write_yaml(pth, obj)

    @staticmethod
    def _parse_dict(obj: dict, validate: bool = True) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        project = obj.get("project")
        kind = obj.get("kind")
        name = build_name(obj.get("name"))
        uuid = build_uuid(obj.get("id"))
        metadata = build_metadata(kind, **obj.get("metadata", {}))
        spec = build_spec(kind, validate=validate, **obj.get("spec", {}))
        status = build_status(kind, **obj.get("status", {}))
        user = obj.get("user")
        return {
            "project": project,
            "name": name,
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
            "user": user,
        }



from __future__ import annotations

import os
import typing

from digitalhub_core.client.builder import check_client_exists, get_client

if typing.TYPE_CHECKING:
    from digitalhub_core.client.objects.dhcore import ClientDHCore


def set_dhcore_env(
    endpoint: str | None = None,
    user: str | None = None,
    password: str | None = None,
    access_token: str | None = None,
    refresh_token: str | None = None,
    client_id: str | None = None,
) -> None:
    """
    Function to set environment variables for DHCore config.
    Note that if the environment variable is already set, it
    will be overwritten. It also ovverides the remote client
    configuration.

    Parameters
    ----------
    endpoint : str
        The endpoint of DHCore.
    user : str
        The user of DHCore.
    password : str
        The password of DHCore.
    access_token : str
        The access token of DHCore.
    refresh_token : str
        The refresh token of DHCore.
    client_id : str
        The client id of DHCore.

    Returns
    -------
    None
    """
    if endpoint is not None:
        os.environ["DHCORE_ENDPOINT"] = endpoint
    if user is not None:
        os.environ["DHCORE_USER"] = user
    if password is not None:
        os.environ["DHCORE_PASSWORD"] = password
    if access_token is not None:
        os.environ["DHCORE_ACCESS_TOKEN"] = access_token
    if refresh_token is not None:
        os.environ["DHCORE_REFRESH_TOKEN"] = refresh_token
    if client_id is not None:
        os.environ["DHCORE_CLIENT_ID"] = client_id

    if check_client_exists(local=False):
        update_client_from_env()


def update_client_from_env() -> None:
    """
    Function to update client from environment variables.

    Returns
    -------
    None
    """
    client: ClientDHCore = get_client(local=False)

    # Update endpoint
    endpoint = os.getenv("DHCORE_ENDPOINT")
    if endpoint is not None:
        client._endpoint = endpoint

    # Update auth

    # If token is set, it will override the other auth options
    access_token = os.getenv("DHCORE_ACCESS_TOKEN")
    refresh_token = os.getenv("DHCORE_REFRESH_TOKEN")
    client_id = os.getenv("DHCORE_CLIENT_ID")

    if access_token is not None:
        if refresh_token is not None:
            client._refresh_token = refresh_token
        if client_id is not None:
            client._client_id = client_id
        client._access_token = access_token
        client._auth_type = "oauth2"
        return

    # Otherwise, if user and password are set, basic auth will be used
    username = os.getenv("DHCORE_USER")
    password = os.getenv("DHCORE_PASSWORD")
    if username is not None and password is not None:
        client._user = username
        client._password = password
        client._auth_type = "basic"


def get_s3_bucket() -> str | None:
    """
    Function to get S3 bucket name.

    Returns
    -------
    str
        The S3 bucket name.
    """
    return os.getenv("S3_BUCKET_NAME", "datalake")



from __future__ import annotations

import typing

from digitalhub_core.entities._base.entity.unversioned import UnversionedEntity
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.entities.run.crud import delete_run, get_run, new_run, run_from_parameters

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities.run.entity import Run
    from digitalhub_core.entities.task.spec import TaskSpec
    from digitalhub_core.entities.task.status import TaskStatus


class Task(UnversionedEntity):
    """
    A class representing a task.
    """

    ENTITY_TYPE = EntityTypes.TASK.value

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: TaskSpec,
        status: TaskStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, uuid, kind, metadata, spec, status, user)
        self.spec: TaskSpec
        self.status: TaskStatus

    ##############################
    #  Task methods
    ##############################

    def run(
        self,
        run_kind: str,
        local_execution: bool = False,
        **kwargs,
    ) -> Run:
        """
        Run task.

        Parameters
        ----------
        run_kind : str
            Kind the object.
        local_execution : bool
            Flag to indicate if the run will be executed locally.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Run
            Run object.
        """
        return self.new_run(
            project=self.project,
            task=self._get_task_string(),
            kind=run_kind,
            local_execution=local_execution,
            **kwargs,
        )

    def _get_task_string(self) -> str:
        """
        Get task string.

        Returns
        -------
        str
            Task string.
        """
        splitted = self.spec.function.split("://")
        return f"{self.kind}://{splitted[1]}"

    ##############################
    # CRUD Methods for Run
    ##############################

    def new_run(self, **kwargs) -> Run:
        """
        Create a new run.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Run
            Run object.
        """
        if kwargs["local_execution"]:
            return run_from_parameters(**kwargs)
        return new_run(**kwargs)

    def get_run(self, entity_key: str) -> Run:
        """
        Get run.

        Parameters
        ----------
        entity_key : str
            Entity key.

        Returns
        -------
        Run
            Run object.
        """
        return get_run(entity_key)

    def delete_run(self, entity_key: str) -> None:
        """
        Delete run.

        Parameters
        ----------
        entity_key : str
            Entity key.

        Returns
        -------
        None
        """
        delete_run(entity_key)



from __future__ import annotations

import typing

from digitalhub_core.entities._base.crud import list_entity_api_ctx
from digitalhub_core.entities._base.entity.executable import ExecutableEntity
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.entities.task.crud import delete_task, task_from_dict, task_from_parameters
from digitalhub_core.runtimes.builder import get_kind_registry
from digitalhub_core.utils.exceptions import BackendError, EntityError

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities.run.entity import Run
    from digitalhub_core.entities.task.entity import Task
    from digitalhub_core.entities.workflow.spec import WorkflowSpec
    from digitalhub_core.entities.workflow.status import WorkflowStatus


class Workflow(ExecutableEntity):
    """
    A class representing a workflow.
    """

    ENTITY_TYPE = EntityTypes.WORKFLOW.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: WorkflowSpec,
        status: WorkflowStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: WorkflowSpec
        self.status: WorkflowStatus

    ##############################
    #  Workflow Methods
    ##############################

    def run(self, **kwargs) -> Run:
        """
        Run workflow.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments passed to Task and Run builders.

        Returns
        -------
        Run
            Run instance.
        """

        # Get kind registry
        kind_reg = get_kind_registry(self.kind)

        # Get task and run kind
        task_kind = kind_reg.get_task_kind_from_action(action="pipeline")
        run_kind = kind_reg.get_run_kind()

        # Create or update new task
        task = self.new_task(task_kind, **kwargs)

        # Raise error if execution is not done by DHCore backend
        if self._context().local:
            raise BackendError("Cannot run workflow with local backend.")

        return task.run(run_kind, local_execution=False, **kwargs)

    def _get_workflow_string(self) -> str:
        """
        Get workflow string.

        Returns
        -------
        str
            Workflow string.
        """
        return super()._get_executable_string()

    def import_tasks(self, tasks: list[dict]) -> None:
        """
        Import tasks from yaml.

        Parameters
        ----------
        tasks : list[dict]
            List of tasks to import.

        Returns
        -------
        None
        """
        # Loop over tasks list, in the case where the function
        # is imported from local file.
        for task in tasks:
            # If task is not a dictionary, skip it
            if not isinstance(task, dict):
                continue

            # Create a new object from dictionary.
            # the form in which tasks are stored in function
            # status
            task_obj = task_from_dict(task)

            # Try to save it in backend to been able to use
            # it for launching runs. In fact, tasks must be
            # persisted in backend to be able to launch runs.
            # Ignore if task already exists
            try:
                task_obj.save()
            except BackendError:
                pass

            # Set task if function is the same. Overwrite
            # status task dict with the new task object
            if task_obj.spec.function == self._get_workflow_string():
                self._tasks[task_obj.kind] = task_obj

    def new_task(self, task_kind: str, **kwargs) -> Task:
        """
        Create new task.

        Parameters
        ----------
        task_kind : str
            Kind the object.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Task
            New task.
        """
        # Override kwargs
        kwargs["project"] = self.project
        kwargs["function"] = self._get_workflow_string()
        kwargs["kind"] = task_kind

        # Create object instance
        task = task_from_parameters(**kwargs)

        exists, task_id = self._check_task_in_backend(task_kind)

        # Save or update task
        if not exists:
            task.save()
        else:
            task.id = task_id
            task.save(update=True)

        self._tasks[task_kind] = task
        return task

    def update_task(self, kind: str, **kwargs) -> None:
        """
        Update task.

        Parameters
        ----------
        kind : str
            Kind the object.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        self._raise_if_not_exists(kind)

        # Update kwargs
        kwargs["project"] = self.project
        kwargs["kind"] = kind
        kwargs["function"] = self._get_workflow_string()
        kwargs["uuid"] = self._tasks[kind].id

        # Update task
        task = task_from_parameters(**kwargs)
        task.save(update=True)
        self._tasks[kind] = task

    def get_task(self, kind: str) -> Task:
        """
        Get task.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        Task
            Task.

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        return self._tasks[kind]

    def delete_task(self, kind: str, cascade: bool = True) -> None:
        """
        Delete task.

        Parameters
        ----------
        kind : str
            Kind the object.
        cascade : bool
            Flag to determine if cascade deletion must be performed.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        delete_task(self._tasks[kind].key, cascade=cascade)
        self._tasks.pop(kind, None)

    def _check_task_in_backend(self, kind: str) -> tuple[bool, str | None]:
        """
        Check if task exists in backend.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        tuple[bool, str | None]
            Flag to determine if task exists in backend and ID if exists.
        """
        # List tasks from backend filtered by function and kind
        params = {"function": self._get_workflow_string(), "kind": kind}
        objs = list_entity_api_ctx(self.project, EntityTypes.TASK.value, params=params)
        try:
            return True, objs[0]["id"]
        except IndexError:
            return False, None

    def _raise_if_not_exists(self, kind: str) -> None:
        """
        Raise error if task is not created.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        if self._tasks.get(kind) is None:
            raise EntityError("Task does not exist.")



"""
Import modules from submodules.
"""
from digitalhub_core.entities.artifact.crud import (
    delete_artifact,
    get_artifact,
    get_artifact_versions,
    import_artifact,
    list_artifacts,
    log_artifact,
    new_artifact,
    update_artifact,
)
from digitalhub_core.entities.function.crud import (
    delete_function,
    get_function,
    get_function_versions,
    import_function,
    list_functions,
    new_function,
    update_function,
)
from digitalhub_core.entities.project.crud import (
    delete_project,
    get_or_create_project,
    get_project,
    import_project,
    load_project,
    new_project,
    update_project,
)
from digitalhub_core.entities.run.crud import delete_run, get_run, import_run, list_runs, new_run, update_run
from digitalhub_core.entities.secret.crud import (
    delete_secret,
    get_secret,
    get_secret_versions,
    import_secret,
    list_secrets,
    new_secret,
    update_secret,
)
from digitalhub_core.entities.task.crud import delete_task, get_task, import_task, list_tasks, new_task, update_task
from digitalhub_core.entities.workflow.crud import (
    delete_workflow,
    get_workflow,
    get_workflow_versions,
    import_workflow,
    list_workflows,
    new_workflow,
    update_workflow,
)
from digitalhub_core.stores.builder import set_store
from digitalhub_core.utils.env_utils import set_dhcore_env



from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj


class Status(ModelObj):
    """
    Base Status class.
    The status class contains information about the state of an entity,
    for example, the state of a RUNNING run, and eventual error message.
    """

    def __init__(self, state: str, message: str | None = None) -> None:
        self.state = state
        self.message = message

    @classmethod
    def from_dict(cls, obj: dict) -> Status:
        """
        Return entity status object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity status.

        Returns
        -------
        EntityStatus
            An entity status object.
        """
        return cls(**obj)



from __future__ import annotations

from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as pq
from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.exceptions import StoreError
from sqlalchemy import MetaData, Table, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.engine.row import LegacyRow
from sqlalchemy.exc import SQLAlchemyError


class SQLStoreConfig(StoreConfig):
    """
    SQL store configuration class.
    """

    host: str
    """SQL host."""

    port: int
    """SQL port."""

    user: str
    """SQL user."""

    password: str
    """SQL password."""

    database: str
    """SQL database name."""


class SqlStore(Store):
    """
    SQL store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on SQL based storage.
    """

    def __init__(self, name: str, store_type: str, config: SQLStoreConfig) -> None:
        super().__init__(name, store_type)
        self.config = config

    ##############################
    # IO methods
    ##############################

    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Download artifacts from storage.

        Parameters
        ----------
        root : str
            The root path of the artifact.
        dst : str
            The destination of the artifact on local filesystem.
        src : list[str]
            List of sources.
        overwrite : bool
            Specify if overwrite existing file(s).

        Returns
        -------
        str
            Destination path of the downloaded artifact.
        """
        # Handle source
        src: str = src[0]

        table_name = self._get_table_name(src) + ".parquet"
        # Case where dst is not provided
        if dst is None:
            dst = Path(self._build_temp("sql")) / table_name
        else:
            self._check_local_dst(str(dst))
            path = Path(dst)

            # Case where dst is a directory
            if path.suffix == "":
                dst = path / table_name

            # Case where dst is a file
            elif path.suffix != ".parquet":
                raise StoreError("The destination path must be a directory or a parquet file.")

            self._check_overwrite(dst, overwrite)
            self._build_path(dst)

        schema = self._get_schema(src)
        table = self._get_table_name(src)
        return self._download_table(schema, table, str(dst))

    def upload(self, src: str | list[str], dst: str | None = None) -> list[tuple[str, str]]:
        """
        Upload an artifact to storage.

        Raises
        ------
        StoreError
            This method is not implemented.
        """
        raise StoreError("SQL store does not support upload.")

    def get_file_info(self, paths: list[str]) -> list[dict]:
        """
        Get file information from SQL based storage.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("SQL store does not support upload.")

    ##############################
    # Private helper methods
    ##############################

    def _get_connection_string(self) -> str:
        """
        Get the connection string.

        Returns
        -------
        str
            The connection string.
        """
        return (
            f"postgresql://{self.config.user}:{self.config.password}@"
            f"{self.config.host}:{self.config.port}/{self.config.database}"
        )

    def _get_engine(self, schema: str | None = None) -> Engine:
        """
        Create engine from connection string.

        Parameters
        ----------
        schema : str
            The schema.

        Returns
        -------
        Engine
            An SQLAlchemy engine.
        """
        connection_string = self._get_connection_string()
        if not isinstance(connection_string, str):
            raise StoreError("Connection string must be a string.")
        try:
            connect_args = {"connect_timeout": 30}
            if schema is not None:
                connect_args["options"] = f"-csearch_path={schema}"
            return create_engine(connection_string, connect_args=connect_args)
        except Exception as ex:
            raise StoreError(f"Something wrong with connection string. Arguments: {str(ex.args)}")

    def _check_factory(self, schema: str | None = None) -> Engine:
        """
        Check if the database is accessible and return the engine.

        Parameters
        ----------
        schema : str
            The schema.

        Returns
        -------
        Engine
            The database engine.
        """
        engine = self._get_engine(schema)
        self._check_access_to_storage(engine)
        return engine

    @staticmethod
    def _parse_path(path: str) -> dict:
        """
        Parse the path and return the components.

        Parameters
        ----------
        path : str
            The path.

        Returns
        -------
        dict
            A dictionary containing the components of the path.
        """
        # Parse path
        err_msg = "Invalid SQL path. Must be sql://<database>/<schema>/<table> or sql://<database>/<table>"
        protocol, pth = path.split("://")
        components = pth.split("/")
        if protocol != "sql" or not (2 <= len(components) <= 3):
            raise ValueError(err_msg)

        # Get components
        database = components[0]
        table = components[-1]
        schema = components[1] if len(components) == 3 else "public"
        return {"database": database, "schema": schema, "table": table}

    def _get_schema(self, uri: str) -> str:
        """
        Get the name of the SQL schema from the URI.

        Parameters
        ----------
        uri : str
            The URI.

        Returns
        -------
        str
            The name of the SQL schema.
        """
        return str(self._parse_path(uri).get("schema"))

    def _get_table_name(self, uri: str) -> str:
        """
        Get the name of the table from the URI.

        Parameters
        ----------
        uri : str
            The URI.

        Returns
        -------
        str
            The name of the table
        """
        return str(self._parse_path(uri).get("table"))

    @staticmethod
    def _check_access_to_storage(engine: Engine) -> None:
        """
        Check if there is access to the storage.

        Parameters
        ----------
        engine : Engine
            An SQLAlchemy engine.

        Returns
        -------
        None

        Raises
        ------
        StoreError
            If there is no access to the storage.
        """
        try:
            engine.connect()
        except SQLAlchemyError:
            engine.dispose()
            raise StoreError("No access to db!")

    def _download_table(self, schema: str, table: str, dst: str) -> str:
        """
        Download a table from SQL based storage.

        Parameters
        ----------
        schema : str
            The origin schema.
        table : str
            The origin table.
        dst : str
            The destination path.

        Returns
        -------
        str
            The destination path.
        """
        engine = self._check_factory(schema=schema)

        # Read the table from the database
        sa_table = Table(table, MetaData(), autoload_with=engine)
        query = sa_table.select()
        with engine.begin() as conn:
            result: list[LegacyRow] = conn.execute(query).fetchall()

        # Parse the result
        data = self._parse_result(result)

        # Convert the result to a pyarrow table and
        # write the pyarrow table to a Parquet file
        arrow_table = pa.Table.from_pydict(data)
        pq.write_table(arrow_table, dst)

        engine.dispose()

        return dst

    @staticmethod
    def _parse_result(result: list[LegacyRow]) -> dict:
        """
        Convert a list of list of tuples to a dict.

        Parameters
        ----------
        result : list[LegacyRow]
            The data to convert.

        Returns
        -------
        dict
            The converted data.
        """
        data_list = [row.items() for row in result]
        data = {}
        for row in data_list:
            for column_name, value in row:
                if column_name not in data:
                    data[column_name] = []
                data[column_name].append(value)
        return data



from __future__ import annotations

from abc import abstractmethod


class Client:
    """
    Base Client class interface.

    The client is an interface that handles the CRUD of an object during
    a session. It manages the creation, reading, updating, deleting and
    listing of objects and comes into two subclasses: Local and DHCore.
    """

    @abstractmethod
    def create_object(self, api: str, obj: dict, **kwargs) -> dict:
        """
        Create object method.
        """

    @abstractmethod
    def read_object(self, api: str, **kwargs) -> dict:
        """
        Read object method.
        """

    @abstractmethod
    def update_object(self, api: str, obj: dict, **kwargs) -> dict:
        """
        Update object method.
        """

    @abstractmethod
    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete object method.
        """

    @abstractmethod
    def list_objects(self, api: str, **kwargs) -> dict:
        """
        Generic call method.
        """

    @staticmethod
    @abstractmethod
    def is_local() -> bool:
        """
        Flag to check if client is local.
        """



from __future__ import annotations

from digitalhub_core.entities._base.spec.base import Spec, SpecParams


class MaterialSpec(Spec):
    """
    Material specification class.
    """

    def __init__(self, path: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.path = path


class MaterialParams(SpecParams):
    """
    Material parameters class.
    """

    path: str
    """Target path to file(s)"""



from __future__ import annotations

from digitalhub_core.utils.exceptions import EntityError
from pydantic import BaseModel


class TaskModel(BaseModel):
    """
    Task model.
    """

    kind: str
    action: str


class RunModel(BaseModel):
    """
    Run model.
    """

    kind: str


class ExecutableModel(BaseModel):
    """
    Executable model.
    """

    kind: str


class DataModel(BaseModel):
    """
    Data model.
    """

    executable: ExecutableModel
    task: list[TaskModel]
    run: RunModel


class KindRegistry:

    """
    Kind registry module for runtimes.
    """

    def __init__(self, data: dict[str, dict[str, str]]) -> None:
        """
        Constructor.

        Parameters
        ----------
        data : dict[str, dict[str, str]]
            Data to validate.

        Examples
        --------
        >>> data = {
        ...     "executable": {
        ...         "kind": "executable-kind"
        ...     },
        ...     "run": {
        ...         "kind": "run-kind"
        ...     },
        ...     "task": [
        ...         {"kind": "task-kind-0", "action": "action-0"},
        ...         {"kind": "task-kind-1", "action": "action-1"},
        ...         {"kind": "task-kind-2", "action": "action-2"},
        ...     ]
        ... }
        """
        validated_data = self._validate_data(data)
        self.data = validated_data

    def _validate_data(self, data: dict[str, dict[str, str]]) -> DataModel:
        """
        Validate data.

        Parameters
        ----------
        data : dict[str, dict[str, str]]
            Data to validate.

        Raises
        ------
        EntityError
            If data is not valid.
        """
        try:
            return DataModel(**data)
        except Exception as e:
            raise EntityError("Invalid data.") from e

    def get_task_kind_from_action(self, action: str) -> str:
        """
        Get task kind from action.

        Parameters
        ----------
        action : str
            Task action.

        Returns
        -------
        str
            Task kind.

        Raises
        ------
        EntityError
            If action is not allowed.
        """
        try:
            for task in self.data.task:
                if task.action == action:
                    return task.kind
        except StopIteration:
            msg = f"Action {action} not allowed."
            raise EntityError(msg)

    def get_action_from_task_kind(self, kind: str) -> str:
        """
        Get action from task.

        Parameters
        ----------
        kind : str
            Task kind.

        Returns
        -------
        str
            Action.

        Raises
        ------
        EntityError
            If task is not allowed.
        """
        try:
            for task in self.data.task:
                if task.kind == kind:
                    return task.action
        except StopIteration:
            msg = f"Task {kind} not allowed."
            raise EntityError(msg)

    def get_run_kind(self) -> str:
        """
        Get run kind.

        Returns
        -------
        str
            Run kind.
        """
        return self.data.run.kind

    def get_executable_kind(self) -> str:
        """
        Get executable kind.

        Returns
        -------
        str
            Executable kind.
        """
        return self.data.executable.kind



from __future__ import annotations

from uuid import uuid4

from pydantic import UUID4, BaseModel


class UUIDValidator(BaseModel):
    """
    Validate UUID format.
    """

    uuid: UUID4


def build_uuid(uuid: str | None = None) -> str:
    """
    Create a uuid if not given. If given, validate it.

    Parameters
    ----------
    uuid : str
        ID of the object (UUID4, e.g. 40f25c4b-d26b-4221-b048-9527aff291e2).

    Returns
    -------
    str
        Validated UUID4.
    """
    if uuid is not None:
        UUIDValidator(uuid=uuid)
        return uuid
    return str(uuid4())



from __future__ import annotations

from enum import Enum


class State(Enum):
    """
    State enumeration.
    """

    BUILT = "BUILT"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    CREATED = "CREATED"
    CREATING = "CREATING"
    DELETED = "DELETED"
    ERROR = "ERROR"
    FSM_ERROR = "FSM_ERROR"
    IDLE = "IDLE"
    NONE = "NONE"
    ONLINE = "ONLINE"
    PENDING = "PENDING"
    READY = "READY"
    RUN_ERROR = "RUN_ERROR"
    RUNNING = "RUNNING"
    STOP = "STOP"
    STOPPED = "STOPPED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"



from __future__ import annotations

from digitalhub_core.entities._base.spec.material import MaterialParams, MaterialSpec


class ArtifactSpec(MaterialSpec):
    """
    Artifact specification.
    """


class ArtifactParams(MaterialParams):
    """
    Artifact base parameters.
    """


class ArtifactSpecArtifact(ArtifactSpec):
    """
    Artifact specification.
    """

    def __init__(
        self,
        path: str,
        src_path: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(path, **kwargs)
        self.src_path = src_path


class ArtifactParamsArtifact(ArtifactParams):
    """
    Artifact parameters.
    """

    src_path: str = None
    """Source path of the artifact."""



from __future__ import annotations

import shutil
from pathlib import Path

from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.file_utils import get_file_info_from_local


class LocalStoreConfig(StoreConfig):
    """
    Local store configuration class.
    """

    path: str
    """Local path."""


class LocalStore(Store):
    """
    Local store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on local filesystem based storage.
    """

    def __init__(self, name: str, store_type: str, config: LocalStoreConfig) -> None:
        super().__init__(name, store_type)
        self.config = config

    ##############################
    # IO methods
    ##############################

    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Download artifacts from storage.

        Parameters
        ----------
        root : str
            The root path of the artifact.
        dst : str
            The destination of the artifact on local filesystem.
        src : list[str]
            List of sources.
        overwrite : bool
            Specify if overwrite existing file(s).

        Returns
        -------
        str
            Destination path of the downloaded artifact.
        """
        raise StoreError("Local store does not support download.")

    def upload(self, src: str | list[str], dst: str | None = None) -> list[tuple[str, str]]:
        """
        Upload an artifact to storage.

        Raises
        ------
        StoreError
            This method is not implemented.
        """
        raise StoreError("Local store does not support upload.")

    def get_file_info(self, paths: list[str]) -> list[dict]:
        """
        Method to get file metadata.

        Parameters
        ----------
        paths : list
            List of source paths.

        Returns
        -------
        list[dict]
            Returns files metadata.
        """
        return [get_file_info_from_local(p) for p in paths]

    ##############################
    # Private I/O methods
    ##############################

    def _get_src_dst_files(self, src: Path, dst: Path) -> list[str]:
        """
        Copy files from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        list[str]
            Returns the list of destination and source paths of the
            copied files.
        """
        return [self._get_src_dst_file(i, dst) for i in src.rglob("*") if i.is_file()]

    def _get_src_dst_file(self, src: Path, dst: Path) -> str:
        """
        Copy file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        str
        """
        dst_pth = self._copy_file(src, dst, True)
        return str(dst_pth), str(src)

    def _copy_dir(self, src: Path, dst: Path, overwrite: bool) -> list[str]:
        """
        Download file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        list[str]
        """
        dst = self._rebuild_path(dst, src)
        shutil.copytree(src, dst, dirs_exist_ok=overwrite)
        return [str(i) for i in dst.rglob("*") if i.is_file()]

    def _copy_file(self, src: Path, dst: Path, overwrite: bool) -> str:
        """
        Copy file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        str
        """
        dst = self._rebuild_path(dst, src)
        self._check_overwrite(dst, overwrite)
        return str(shutil.copy2(src, dst))

    def _rebuild_path(self, dst: Path, src: Path) -> Path:
        """
        Rebuild path.

        Parameters
        ----------
        dst : Path
            The destination path.
        src : Path
            The source path.

        Returns
        -------
        Path
            The rebuilt path.
        """
        if dst.is_dir():
            if src.is_absolute():
                raise StoreError("Source must be a relative path if the destination is a directory.")
            dst = dst / src
        self._build_path(dst)
        return dst

    ##############################
    # Static methods
    ##############################

    @staticmethod
    def is_partition_or_dir(path: str) -> bool:
        """
        Check if path is a directory or a partition.

        Parameters
        ----------
        path : str
            The path to check.

        Returns
        -------
        bool
        """
        return Path(path).is_dir()

    @staticmethod
    def build_object_path(root: str, paths: str | list[str]) -> list[str]:
        """
        Method to build object path.

        Parameters
        ----------
        root : str
            The root of the object path.
        paths : str | list[str]
            The path to build.

        Returns
        -------
        list[str]
            Returns the path of the object.
        """
        if isinstance(paths, str):
            paths = [paths]
        return [str(Path(root) / path) for path in paths]



from __future__ import annotations

import os
import shutil
import warnings
from pathlib import Path
from urllib.parse import urlparse

try:
    from git import Repo
except ImportError as e:
    if "Bad git executable." in e.args[0]:
        warnings.warn("git is not installed. Please install git and try again.", RuntimeWarning)


def clone_repository(url: str, path: Path) -> None:
    """
    Clone git repository.

    Parameters
    ----------
    url : str
        URL of the repository.
    path : Path
        Path where to save the repository.

    Returns
    -------
    None
    """
    clean_path(path)
    checkout_object = get_checkout_object(url)
    url = add_credentials_git_remote_url(url)
    repo = clone_from_url(url, path)
    if checkout_object != "":
        repo.git.checkout(checkout_object)


def get_checkout_object(url: str) -> str:
    """
    Get checkout object from url fragment.

    Parameters
    ----------
    url : str
        URL of the repository.

    Returns
    -------
    str
        Checkout object (branch, tag, commit).
    """
    return urlparse(url).fragment


def clean_path(path: Path) -> None:
    """
    Clean path from any files.

    Parameters
    ----------
    path : Path

    Returns
    -------
    None
    """

    shutil.rmtree(path, ignore_errors=True)


def get_git_username_password_from_token(token: str) -> tuple[str, str]:
    """
    Parse token to get username and password. The token
    can be one of the following:

    - GitHub/GitLab personal access token (github_pat_.../glpat...)
    - GitHub/GitLab access token
    - Other generic token

    Parameters
    ----------
    token : str
        Token to parse.

    Returns
    -------
    tuple[str, str]
        Username and password.
    """
    # Mutued from mlrun
    if token.startswith("github_pat_") or token.startswith("glpat"):
        username = "oauth2"
        password = token
    else:
        username = token
        password = "x-oauth-basic"
    return username, password


def add_credentials_git_remote_url(url: str) -> str:
    """
    Add credentials to git remote url.

    Parameters
    ----------
    url : str
        URL of the repository.

    Returns
    -------
    str
        URL with credentials.
    """
    url_obj = urlparse(url)

    # Get credentials from environment variables
    username = os.getenv("GIT_USERNAME")
    password = os.getenv("GIT_PASSWORD")
    token = os.getenv("GIT_TOKEN")

    # Get credentials from token. Override username and password
    if token is not None:
        username, password = get_git_username_password_from_token(token)

    # Add credentials to url if needed
    if username is not None and password is not None:
        return f"https://{username}:{password}@{url_obj.hostname}{url_obj.path}"
    return f"https://{url_obj.hostname}{url_obj.path}"


def clone_from_url(url: str, path: Path) -> Repo:
    """
    Clone repository from url. Wraps git.Repo.clone_from.

    Parameters
    ----------
    url : str
        HTTP(S) URL of the repository.
    path : Path
        Path where to save the repository.

    Returns
    -------
    Repo
        Cloned repository.
    """
    return Repo.clone_from(url=url, to_path=path)



from __future__ import annotations

import os
import typing

from digitalhub_core.stores.objects.base import StoreParameters
from digitalhub_core.stores.objects.local import LocalStore, LocalStoreConfig
from digitalhub_core.stores.objects.remote import RemoteStore, RemoteStoreConfig
from digitalhub_core.stores.objects.s3 import S3Store, S3StoreConfig
from digitalhub_core.stores.objects.sql import SqlStore, SQLStoreConfig
from digitalhub_core.utils.exceptions import StoreError
from digitalhub_core.utils.uri_utils import map_uri_scheme
from pydantic import ValidationError

if typing.TYPE_CHECKING:
    from digitalhub_core.stores.objects.base import Store


REGISTRY_STORES = {
    "local": LocalStore,
    "s3": S3Store,
    "remote": RemoteStore,
    "sql": SqlStore,
}


class StoreBuilder:
    """
    Store builder class.
    """

    def __init__(self) -> None:
        self._instances: dict[str, Store] = {}
        self._default: Store | None = None
        self._def_scheme = "s3"

    def build(self, store_cfg: StoreParameters) -> None:
        """
        Build a store instance and register it.
        It overrides any existing instance.

        Parameters
        ----------
        store_cfg : StoreParameters
            Store configuration.

        Returns
        -------
        None
        """
        scheme = map_uri_scheme(store_cfg.type)
        self._instances[scheme] = self.build_store(store_cfg)

    def get(self, uri: str) -> Store:
        """
        Get a store instance by URI.

        Parameters
        ----------
        uri : str
            URI to parse.

        Returns
        -------
        Store
            The store instance.
        """
        scheme = map_uri_scheme(uri)
        if scheme not in self._instances:
            store_cfg = get_env_store_config(scheme)
            self._instances[scheme] = self.build_store(store_cfg)
        return self._instances[scheme]

    def default(self) -> Store:
        """
        Get the default store instance.

        Returns
        -------
        Store
            The default store instance.

        Raises
        ------
        StoreError
            If no default store is set.
        """
        if self._default is None:
            store_cfg = get_env_store_config(self._def_scheme)
            self._default = self.build_store(store_cfg)
        return self._default

    def build_store(self, cfg: StoreParameters) -> Store:
        """
        Build a store instance.

        Parameters
        ----------
        cfg : StoreParameters
            Store configuration.

        Returns
        -------
        Store
            The store instance.

        Raises
        ------
        NotImplementedError
            If the store type is not implemented.
        """
        try:
            obj = REGISTRY_STORES[cfg.type](cfg.name, cfg.type, cfg.config)
            if cfg.is_default and self._default is not None:
                raise StoreError("Only one default store!")
            return obj
        except KeyError as e:
            raise NotImplementedError from e

    @staticmethod
    def _check_config(config: StoreParameters | dict) -> StoreParameters:
        """
        Check the store configuration validity.

        Parameters
        ----------
        config : StoreParameters | dict
            The store configuration.

        Returns
        -------
        StoreParameters
            The store configuration.

        Raises
        ------
        TypeError
            If the config parameter is not a StoreParameters instance or a well-formed dictionary.
        """
        if not isinstance(config, StoreParameters):
            try:
                return StoreParameters(**config)
            except TypeError as e:
                raise StoreError("Invalid store configuration type.") from e
            except ValidationError as e:
                raise StoreError("Malformed store configuration parameters.") from e
        return config


def get_env_store_config(scheme: str) -> StoreParameters:
    """
    Get a store configuration from the environment.

    Parameters
    ----------
    scheme : str
        URI scheme.

    Returns
    -------
    StoreParameters
        The store configuration based on the scheme.

    Raises
    ------
    ValueError
        If the scheme is not supported.
    """
    if scheme == "s3":
        return StoreParameters(
            name="s3",
            type="s3",
            config=S3StoreConfig(
                endpoint_url=os.getenv("S3_ENDPOINT_URL"),  # type: ignore
                aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),  # type: ignore
                aws_secret_access_key=os.getenv("AWS_SECRET_ACCESS_KEY"),  # type: ignore
                bucket_name=os.getenv("S3_BUCKET_NAME"),  # type: ignore
            ),
        )
    if scheme == "sql":
        return StoreParameters(
            name="sql",
            type="sql",
            config=SQLStoreConfig(
                host=os.getenv("POSTGRES_HOST"),  # type: ignore
                port=os.getenv("POSTGRES_PORT"),  # type: ignore
                user=os.getenv("POSTGRES_USER"),  # type: ignore
                password=os.getenv("POSTGRES_PASSWORD"),  # type: ignore
                database=os.getenv("POSTGRES_DATABASE"),  # type: ignore
                pg_schema=os.getenv("POSTGRES_SCHEMA"),  # type: ignore
            ),
        )
    if scheme == "remote":
        return StoreParameters(
            name="remote",
            type="remote",
            config=RemoteStoreConfig(),
        )
    if scheme == "local":
        return StoreParameters(
            name="local",
            type="local",
            config=LocalStoreConfig(
                path="tempsdk",
            ),
        )
    raise ValueError(f"Unsupported scheme {scheme}")


def set_store(store_cfg: StoreParameters) -> None:
    """
    Set a new store instance with the given configuration.

    Parameters
    ----------
    store_cfg : StoreParameters
        Store configuration.

    Returns
    -------
    None
    """
    store_builder.build(store_cfg)


def get_store(uri: str) -> Store:
    """
    Get store instance by uri.

    Parameters
    ---------
    uri : str
        URI to parse.

    Returns
    -------
    Store
        Store instance.
    """
    return store_builder.get(uri)


def get_default_store() -> Store:
    """
    Get the default store instance. The default store is the one that
    can persist artifacts and dataitems.

    Returns
    -------
    Store
        Default store instance.
    """
    return store_builder.default()


store_builder = StoreBuilder()



from __future__ import annotations

import typing

from digitalhub_core.context.builder import check_context
from digitalhub_core.entities._base.crud import delete_entity_api_ctx, list_entity_api_ctx, read_entity_api_ctx
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.entities.task.builder import task_from_dict, task_from_parameters
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.task.entity import Task


ENTITY_TYPE = EntityTypes.TASK.value


def new_task(
    project: str,
    kind: str,
    uuid: str | None = None,
    labels: list[str] | None = None,
    function: str | None = None,
    **kwargs,
) -> Task:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object (UUID4, e.g. 40f25c4b-d26b-4221-b048-9527aff291e2).
    labels : list[str]
        List of labels.
    function : str
        Name of the executable associated with the task.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Task
        Object instance.

    Examples
    --------
    >>> obj = new_function(project="my-project",
    >>>                    name="my-function",
    >>>                    kind="python+task",
    >>>                    task="task-string"
    """
    check_context(project)
    obj = task_from_parameters(
        project=project,
        kind=kind,
        uuid=uuid,
        labels=labels,
        function=function,
        **kwargs,
    )
    obj.save()
    return obj


def get_task(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> Task:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity ID.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Task
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_task("store://my-task-key")

    Using entity ID:
    >>> obj = get_task("my-task-id"
    >>>               project="my-project")
    """
    if not identifier.startswith("store://"):
        raise EntityError("Task has no name. Use key instead.")
    obj = read_entity_api_ctx(
        identifier,
        ENTITY_TYPE,
        project=project,
        entity_id=identifier,
        **kwargs,
    )
    return task_from_dict(obj)


def list_tasks(project: str, **kwargs) -> list[Task]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Task]
        List of object instances.

    Examples
    --------
    >>> objs = list_tasks(project="my-project")
    """
    objs = list_entity_api_ctx(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )
    return [task_from_dict(obj) for obj in objs]


def import_task(file: str) -> Task:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Task
        Object instance.

    Example
    -------
    >>> obj = import_task("my-task.yaml")
    """
    obj: dict = read_yaml(file)
    return task_from_dict(obj)


def update_task(entity: Task) -> Task:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Task
        Object to update.

    Returns
    -------
    Task
        Entity updated.

    Examples
    --------
    >>> obj = update_task(obj)
    """
    return entity.save(update=True)


def delete_task(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_task("store://my-task-key")

    Otherwise:
    >>> obj = delete_task("task-name",
    >>>                  project="my-project",
    >>>                  delete_all_versions=True)
    """
    if not identifier.startswith("store://"):
        raise EntityError("Task has no name. Use key instead.")
    return delete_entity_api_ctx(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
        **kwargs,
    )



from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.api import (
    api_base_create,
    api_base_delete,
    api_base_list,
    api_base_read,
    api_base_update,
    api_ctx_create,
    api_ctx_data,
    api_ctx_delete,
    api_ctx_files,
    api_ctx_list,
    api_ctx_logs,
    api_ctx_read,
    api_ctx_stop,
    api_ctx_update,
)
from digitalhub_core.entities.utils import parse_entity_key

if typing.TYPE_CHECKING:
    from digitalhub_core.client.objects.base import Client


def create_entity_api_base(
    client: Client,
    entity_type: str,
    entity_dict: dict,
    **kwargs,
) -> dict:
    """
    Create object in backend.

    Parameters
    ----------
    client : Client
        Client instance.
    entity_type : str
        Entity type.
    entity_dict : dict
        Object instance.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    api = api_base_create(entity_type)
    return client.create_object(api, entity_dict, **kwargs)


def read_entity_api_base(
    client: Client,
    entity_type: str,
    entity_name: str,
    **kwargs,
) -> dict:
    """
    Read object from backend.

    Parameters
    ----------
    client : Client
        Client instance.
    entity_type : str
        Entity type.
    entity_name : str
        Entity name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    api = api_base_read(entity_type, entity_name)
    return client.read_object(api, **kwargs)


def update_entity_api_base(
    client: Client,
    entity_type: str,
    entity_name: str,
    entity_dict: dict,
    **kwargs,
) -> dict:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    client : Client
        Client instance.
    entity_type : str
        Entity type.
    entity_name : str
        Entity name.
    entity_dict : dict
        Object instance.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    api = api_base_update(entity_type, entity_name)
    return client.update_object(api, entity_dict, **kwargs)


def delete_entity_api_base(
    client: Client,
    entity_type: str,
    entity_name: str,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    client : Client
        Client instance.
    entity_type : str
        Entity type.
    entity_name : str
        Entity name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    if "params" not in kwargs:
        kwargs["params"] = {}
    if "cascade" in kwargs["params"]:
        kwargs["params"]["cascade"] = str(kwargs["cascade"]).lower()
    if "cascade" in kwargs:
        kwargs["params"]["cascade"] = str(kwargs.pop("cascade")).lower()
    api = api_base_delete(entity_type, entity_name)
    return client.delete_object(api, **kwargs)


def list_entity_api_base(
    client: Client,
    entity_type: str,
    **kwargs,
) -> list[dict]:
    """
    List objects from backend.

    Parameters
    ----------
    entity_type : str
        Entity type.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[dict]
        List of objects.
    """
    api = api_base_list(entity_type)
    return client.list_objects(api, **kwargs)


def create_entity_api_ctx(
    project: str,
    entity_type: str,
    entity_dict: dict,
    **kwargs,
) -> dict:
    """
    Create object in backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_dict : dict
        Object instance.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    api = api_ctx_create(project, entity_type)
    return get_context(project).create_object(api, entity_dict, **kwargs)


def read_entity_api_ctx(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> dict:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """

    if identifier.startswith("store://"):
        project, entity_type, _, _, entity_id = parse_entity_key(identifier)
        api = api_ctx_read(project, entity_type, entity_id)
        return get_context(project).read_object(api, **kwargs)

    if project is None or entity_type is None:
        raise ValueError("Project and entity type must be specified.")

    if "params" not in kwargs:
        kwargs["params"] = {}

    if entity_id is None:
        kwargs["params"]["name"] = identifier
        api = api_ctx_list(project, entity_type)
        return get_context(project).list_objects(api, **kwargs)[0]

    api = api_ctx_read(project, entity_type, entity_id)
    return get_context(project).read_object(api, **kwargs)


def read_entity_api_ctx_versions(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    **kwargs,
) -> list[dict]:
    """
    Get all versions object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[dict]
        Object instances.
    """
    if not identifier.startswith("store://"):
        if project is None or entity_type is None:
            raise ValueError("Project and entity type must be specified.")
        entity_name = identifier
    else:
        project, entity_type, _, entity_name, _ = parse_entity_key(identifier)

    if "params" not in kwargs:
        kwargs["params"] = {}
    kwargs["params"]["name"] = entity_name
    kwargs["params"]["versions"] = "all"

    api = api_ctx_list(project, entity_type)
    return get_context(project).list_objects(api, **kwargs)


def update_entity_api_ctx(
    project: str,
    entity_type: str,
    entity_id: str,
    entity_dict: dict,
    **kwargs,
) -> dict:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.
    entity_dict : dict
        Entity dictionary.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(project, entity_type, entity_id=entity_id)
    return get_context(project).update_object(api, entity_dict, **kwargs)


def delete_entity_api_ctx(
    identifier: str,
    entity_type: str,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    if "params" not in kwargs:
        kwargs["params"] = {}
    if "cascade" in kwargs["params"]:
        kwargs["params"]["cascade"] = str(kwargs["cascade"]).lower()
    if "cascade" in kwargs:
        kwargs["params"]["cascade"] = str(kwargs.pop("cascade")).lower()

    delete_all_versions: bool = kwargs.pop("delete_all_versions", False)

    if identifier.startswith("store://"):
        project, _, _, _, entity_id = parse_entity_key(identifier)
        api = api_ctx_delete(project, entity_type, entity_id)
        return get_context(project).delete_object(api, **kwargs)

    if project is None:
        raise ValueError("Project must be provided.")

    if entity_id is not None:
        api = api_ctx_delete(project, entity_type, entity_id)
    else:
        kwargs["params"]["name"] = identifier
        api = api_ctx_list(project, entity_type)
        if delete_all_versions:
            return get_context(project).delete_object(api, **kwargs)
        obj = get_context(project).list_objects(api, **kwargs)[0]
        entity_id = obj["id"]

    api = api_ctx_delete(project, entity_type, entity_id)
    return get_context(project).delete_object(api, **kwargs)


def list_entity_api_ctx(
    project: str,
    entity_type: str,
    **kwargs,
) -> list[dict]:
    """
    List objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[dict]
        List of objects.
    """
    api = api_ctx_list(project, entity_type)
    return get_context(project).list_objects(api, **kwargs)


def set_data_api(
    project: str,
    entity_type: str,
    data: dict,
    **kwargs,
) -> None:
    """
    Set data in backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    data : dict
        Data dictionary.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    None
    """
    api = api_ctx_data(project, entity_type)
    get_context(project).update_object(api, data, **kwargs)


def get_data_api(
    project: str,
    entity_type: str,
    **kwargs,
) -> dict:
    """
    Read data from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_data(project, entity_type)
    return get_context(project).read_object(api, **kwargs)


def logs_api(
    project: str,
    entity_type: str,
    entity_id: str,
    **kwargs,
) -> dict:
    """
    Read logs from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_logs(project, entity_type, entity_id)
    return get_context(project).read_object(api, **kwargs)


def stop_api(
    project: str,
    entity_type: str,
    entity_id: str,
    **kwargs,
) -> None:
    """
    Stop object in backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    None
    """
    api = api_ctx_stop(project, entity_type, entity_id)
    get_context(project).create_object(api, obj={}, **kwargs)


def files_info_get_api(
    project: str,
    entity_type: str,
    entity_id: str,
    **kwargs,
) -> list[dict]:
    """
    Get files info from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[dict]
        Response from backend.
    """
    api = api_ctx_files(project, entity_type, entity_id)
    return get_context(project).read_object(api, **kwargs)


def files_info_put_api(
    project: str,
    entity_type: str,
    entity_id: str,
    entity_list: list[dict],
    **kwargs,
) -> None:
    """
    Get files info from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.
    entity_list : list[dict]
        Entity list.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    None
    """
    api = api_ctx_files(project, entity_type, entity_id)
    get_context(project).update_object(api, entity_list, **kwargs)



from __future__ import annotations

import logging

# Create logger
LOGGER = logging.getLogger("digitalhub-core")
LOGGER.setLevel(logging.INFO)

# Create formatter
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

# Create console handler and set formatter
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)

# Set console handler to the logger
LOGGER.addHandler(console_handler)



from __future__ import annotations

import typing

from digitalhub_core.registry.registry import registry
from digitalhub_core.registry.utils import import_class
from digitalhub_core.utils.generic_utils import get_timestamp

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.registry.models import RegistryEntry


def build_metadata(kind: str, **kwargs) -> Metadata:
    """
    Build entity metadata object. The builder takes as input
    the kind of metadata's object to build and the keyword
    arguments to pass to the metadata's constructor.
    The specific Metadata class is searched in the global
    registry, where lies info about where to find the class.
    The arguments are parsed, eventually adding default values,
    and then passed to the constructor.

    Parameters
    ----------
    kind : str
        Registry entry kind.
    **kwargs : dict
        Keyword arguments for the constructor.

    Returns
    -------
    Metadata
        Metadata object.
    """
    infos: RegistryEntry = getattr(registry, kind)
    metadata = import_class(infos.metadata.module, infos.metadata.class_name)
    kwargs = parse_arguments(**kwargs)
    return metadata(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values if necessary.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    dict
        A dictionary containing the entity metadata attributes.
    """
    if "created" not in kwargs or kwargs["created"] is None:
        kwargs["created"] = get_timestamp()
    if "updated" not in kwargs or kwargs["updated"] is None:
        kwargs["updated"] = kwargs["created"]
    return kwargs



from __future__ import annotations

import importlib
import importlib.metadata
import pkgutil
import re
from types import ModuleType


def import_module(package: str) -> ModuleType:
    """
    Import modules from package name.

    Parameters
    ----------
    package : str
        Package name.

    Returns
    -------
    ModuleType
        Module.
    """
    try:
        return importlib.import_module(package)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"Package {package} not found.")
    except Exception as e:
        raise e


def import_class(module_to_import: str, class_name: str) -> type:
    """
    Import class from implemented module.

    Parameters
    ----------
    module_to_import : str
        Module name.
    class_name : str
        Class name.

    Returns
    -------
    type
        Class.
    """
    module = import_module(module_to_import)
    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ModuleNotFoundError(f"Module {module_to_import} has no '{class_name}' class.")


def list_runtimes() -> list[str]:
    """
    List installed runtimes for digitalhub.

    Returns
    -------
    list
        List of installed runtimes names.
    """
    pattern = r"digitalhub_runtime_.*"
    runtimes = []
    try:
        for _, name, _ in pkgutil.iter_modules():
            if re.match(pattern, name):
                runtimes.append(name)
        return runtimes
    except Exception:
        raise RuntimeError("Error listing installed runtimes.")


def register_runtimes_entities() -> None:
    """
    Register runtimes and related entities into registry.

    Returns
    -------
    None
    """
    for package in list_runtimes():
        import_module(package)


def register_layer_entities() -> None:
    """
    Register layer and related entities into registry.

    Returns
    -------
    None
    """
    for layer in ["core", "data", "ml", "ai"]:
        # Check if package exists
        package = f"digitalhub_{layer}"
        try:
            importlib.metadata.distribution(package)
        except importlib.metadata.PackageNotFoundError:
            # Return because the layers are pyramidal from
            # core to ai
            return

        # Try ot import registry from entities.registries module
        try:
            import_module(f"{package}.entities.registries")
        except Exception:
            pass


def create_info(
    root: str,
    entity_type: str,
    prefix: str,
    suffix: str | None = "",
    runtime_info: dict | None = None,
) -> dict:
    """
    Create entity info.

    Parameters
    ----------
    root : str
        Root module.
    entity_type : str
        Entity type.
    prefix : str
        Entity prefix.
    suffix : str
        Entity suffix.
    runtime_info : dict
        Runtime info.

    Returns
    -------
    dict
        Entity info.
    """
    dict_ = {
        "entity_type": entity_type,
        "spec": {
            "module": f"{root}.{entity_type}.spec",
            "class_name": f"{prefix}Spec{suffix}",
            "parameters_validator": f"{prefix}Params{suffix}",
        },
        "status": {
            "module": f"{root}.{entity_type}.status",
            "class_name": f"{prefix}Status{suffix}",
        },
        "metadata": {
            "module": "digitalhub_core.entities._base.metadata",
            "class_name": "Metadata",
        },
    }
    # Add runtime only if provided
    # (in functions, tasks, runs and workflows)
    if runtime_info is not None:
        dict_["runtime"] = runtime_info
    return dict_



from __future__ import annotations

from digitalhub_core.entities._base.spec.base import Spec, SpecParams


class ProjectSpec(Spec):
    """
    Project specification.
    """

    def __init__(
        self,
        context: str | None = None,
        functions: list | None = None,
        artifacts: list | None = None,
        workflows: list | None = None,
        **kwargs,
    ) -> None:
        """
        Initializes a new instance of the ProjectSpec class.

        Parameters
        ----------
        context : str
            The project's context folder.
        functions : list
            List of project's functions.
        artifacts : list
            List of project's artifacts.
        workflows : list
            List of project's workflows.
        """
        self.context = context if context is not None else "./"
        self.functions = functions if functions is not None else []
        self.artifacts = artifacts if artifacts is not None else []
        self.workflows = workflows if workflows is not None else []


class ProjectParams(SpecParams):
    """
    Parameters model for project.
    """

    context: str = None
    """The project's context."""

    functions: list = None
    """List of project's functions."""

    artifacts: list = None
    """List of project's artifacts."""

    workflows: list = None
    """List of project's workflows."""



from __future__ import annotations

import typing
from pathlib import Path

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.project.entity._base import Project


class Context:
    """
    Context class built forom a `Project` instance. It contains
    some information about the project, such as the project name,
    a client instance (local or non-local), the local context
    project path and information about client locality.
    It exposes CRUD operations for the entities and act as a layer
    between the project object and its client.
    """

    def __init__(self, project: Project) -> None:
        self.name = project.name
        self.client = project._client
        self.local = project._client.is_local()
        self.root = Path(project.spec.context)

    def create_object(self, api: str, obj: dict, **kwargs) -> dict:
        """
        Create an object.

        Parameters
        ----------
        api : str
            Create API.
        obj : dict
            Object to create.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            Response object.
        """
        return self.client.create_object(api, obj, **kwargs)

    def read_object(self, api: str, **kwargs) -> dict:
        """
        Read an object.

        Parameters
        ----------
        api : str
            Read API.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            Response object.
        """
        return self.client.read_object(api, **kwargs)

    def update_object(self, api: str, obj: dict, **kwargs) -> dict:
        """
        Update an object.

        Parameters
        ----------
        api : str
            Update API.
        obj : dict
            Object to update.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            Response object.
        """
        return self.client.update_object(api, obj, **kwargs)

    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            Delete API.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            Response object.
        """
        return self.client.delete_object(api, **kwargs)

    def list_objects(self, api: str, **kwargs) -> dict:
        """
        List objects.

        Parameters
        ----------
        api : str
            The api to list the objects with.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            The list of objects.
        """
        return self.client.list_objects(api, **kwargs)



from __future__ import annotations

import typing
from pathlib import Path

from digitalhub_core.client.builder import get_client
from digitalhub_core.context.builder import set_context
from digitalhub_core.entities._base.crud import (
    create_entity_api_base,
    read_entity_api_base,
    read_entity_api_ctx,
    update_entity_api_base,
)
from digitalhub_core.entities._base.entity.base import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.name import build_name
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.entity_types import EntityTypes
from digitalhub_core.utils.exceptions import BackendError
from digitalhub_core.utils.generic_utils import get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata
    from digitalhub_core.entities.project.spec import ProjectSpec
    from digitalhub_core.entities.project.status import ProjectStatus

CTX_ENTITIES = []
FUNC_MAP = {}


class Project(Entity):
    """
    A class representing a project.
    """

    ENTITY_TYPE = EntityTypes.PROJECT.value

    def __init__(
        self,
        name: str,
        kind: str,
        metadata: Metadata,
        spec: ProjectSpec,
        status: ProjectStatus,
        user: str | None = None,
        local: bool = False,
    ) -> None:
        super().__init__(kind, metadata, spec, status, user)
        self.id = name
        self.name = name
        self.key = f"store://{name}"
        self.spec: ProjectSpec
        self.status: ProjectStatus

        # Add attributes to be used in the to_dict method
        self._obj_attr.extend(["id", "name"])

        # Set client
        self._client = get_client(local)

        # Set context
        set_context(self)

    ##############################
    #  Save / Refresh / Export
    ##############################

    def save(self, update: bool = False) -> Project:
        """
        Save entity into backend.

        Parameters
        ----------
        update : bool
            If True, the object will be updated.

        Returns
        -------
        Project
            Entity saved.
        """
        obj = self._refresh_to_dict()

        if not update:
            new_obj = create_entity_api_base(self._client, self.ENTITY_TYPE, obj)
            new_obj["local"] = self._client.is_local()
            self._update_attributes(new_obj)
            return self

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        new_obj = update_entity_api_base(self._client, self.ENTITY_TYPE, obj)
        new_obj["local"] = self._client.is_local()
        self._update_attributes(new_obj)
        return self

    def refresh(self) -> Project:
        """
        Refresh object from backend.

        Returns
        -------
        Project
            Project object.
        """
        new_obj = read_entity_api_base(self._client, self.ENTITY_TYPE, self.name)
        new_obj["local"] = self._client.is_local()
        self._update_attributes(new_obj)
        return self

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file. If the objects are not embedded, the objects are
        exported as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self._refresh_to_dict()

        if filename is None:
            filename = f"{self.kind}_{self.name}.yml"
        pth = Path(self.spec.context) / filename
        pth.parent.mkdir(parents=True, exist_ok=True)
        write_yaml(pth, obj)

        for entity_type in CTX_ENTITIES:
            entity_list = obj.get("spec", {}).get(entity_type, [])
            if not entity_list:
                continue
            self._export_not_embedded(entity_list, entity_type)

    def _refresh_to_dict(self) -> dict:
        """
        Try to refresh object to collect entities related to project.

        Returns
        -------
        dict
            Entity object in dictionary format.
        """
        try:
            return self.refresh().to_dict()
        except BackendError:
            return self.to_dict()

    def _export_not_embedded(self, entity_list: list, entity_type: str) -> None:
        """
        Export project objects if not embedded.

        Parameters
        ----------
        entity_list : list
            Entity list.

        Returns
        -------
        None
        """
        for entity in entity_list:
            if not entity["metadata"]["embedded"]:
                obj: dict = read_entity_api_ctx(entity["key"])
                ent = FUNC_MAP[entity_type](obj)
                ent.export()

    ##############################
    #  Static interface methods
    ##############################

    @staticmethod
    def _parse_dict(obj: dict, validate: bool = True) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        name = build_name(obj.get("name"))
        kind = obj.get("kind")
        metadata = build_metadata(kind, **obj.get("metadata", {}))
        spec = build_spec(kind, validate=validate, **obj.get("spec", {}))
        status = build_status(kind, **obj.get("status", {}))
        user = obj.get("user")
        local = obj.get("local", False)
        return {
            "name": name,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
            "user": user,
            "local": local,
        }



from __future__ import annotations

from urllib.parse import urlparse


def map_uri_scheme(uri: str) -> str:
    """
    Map an URI scheme to a common scheme.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        Mapped scheme type.

    Raises
    ------
    ValueError
        If the scheme is unknown.
    """
    scheme = urlparse(uri).scheme
    if scheme in [""]:
        return "local"
    if scheme in ["file", "local"]:
        raise ValueError("For local path, do not use any scheme")
    if scheme in ["http", "https"]:
        return "remote"
    if scheme in ["s3", "s3a", "s3n", "zip+s3"]:
        return "s3"
    if scheme in ["sql", "postgresql"]:
        return "sql"
    if scheme in ["git", "git+http", "git+https"]:
        return "git"
    raise ValueError(f"Unknown scheme '{scheme}'!")


def check_local_path(path: str) -> bool:
    """
    Check if path is local.

    Parameters
    ----------
    path : str
        Path of some source.

    Returns
    -------
    bool
        True if path is local.
    """
    scheme = map_uri_scheme(path)
    return scheme == "local"



from __future__ import annotations

import typing

from digitalhub_core.registry.registry import registry
from digitalhub_core.registry.utils import import_class

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.spec.base import Spec
    from digitalhub_core.registry.models import RegistryEntry


def build_spec(kind: str, validate: bool = True, **kwargs) -> Spec:
    """
    Build entity spec object. The builder takes as input
    the kind of spec's object to build, a validation flag
    and the keyword arguments to pass to the spec's constructor.
    The specific Spec class is searched in the global registry,
    where lies info about where to find the class.
    If the validation flag is set to True, the arguments are
    validated against a pydantic schema and then passed to the
    constructor.

    Parameters
    ----------
    kind : str
        Registry entry kind.
    validate : bool
        Flag to determine if validate kwargs.
    **kwargs : dict
        Keyword arguments for the constructor.

    Returns
    -------
    Spec
        Spec object.
    """
    infos: RegistryEntry = getattr(registry, kind)
    spec = import_class(infos.spec.module, infos.spec.class_name)
    if validate:
        validator = import_class(infos.spec.module, infos.spec.parameters_validator)
        kwargs = validator(**kwargs).dict(by_alias=True, exclude_none=True)
    return spec(**kwargs)



from __future__ import annotations

import base64
import json
import os
from datetime import datetime
from pathlib import Path
from typing import Any
from urllib.parse import urlparse
from zipfile import ZipFile

import numpy as np
from boto3 import client as boto3_client
from digitalhub_core.utils.io_utils import read_text
from requests import get as requests_get


def get_timestamp() -> str:
    """
    Get the current timestamp timezoned.

    Returns
    -------
    str
        The current timestamp.
    """
    return datetime.now().astimezone().isoformat()


def decode_string(string: str) -> str:
    """
    Decode a string from base64.

    Parameters
    ----------
    string : str
        The string to decode.

    Returns
    -------
    str
        The string decoded from base64.
    """
    return base64.b64decode(string).decode()


def encode_string(string: str) -> str:
    """
    Encode a string in base64.

    Parameters
    ----------
    string : str
        The string to encode.

    Returns
    -------
    str
        The string encoded in base64.
    """
    return base64.b64encode(string.encode()).decode()


def encode_source(path: str) -> str:
    """
    Read a file and encode in base64 the content.

    Parameters
    ----------
    path : str
        The file path to read.

    Returns
    -------
    str
        The file content encoded in base64.
    """
    return encode_string(read_text(path))


def requests_chunk_download(source: str, filename: Path) -> None:
    """
    Download a file in chunks.

    Parameters
    ----------
    source : str
        URL to download the file.
    filename : Path
        Path where to save the file.

    Returns
    -------
    None
    """
    with requests_get(source, stream=True) as r:
        r.raise_for_status()
        with filename.open("wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)


def extract_archive(path: Path, filename: Path) -> None:
    """
    Extract a zip archive.

    Parameters
    ----------
    path : Path
        Path where to extract the archive.
    filename : Path
        Path to the archive.

    Returns
    -------
    None
    """
    with ZipFile(filename, "r") as zip_file:
        zip_file.extractall(path)


def get_s3_source(bucket: str, key: str, filename: Path) -> None:
    """
    Get S3 source.

    Parameters
    ----------
    bucket : str
        S3 bucket name.
    key : str
        S3 object key.
    filename : Path
        Path where to save the function source.

    Returns
    -------
    None
    """
    s3 = boto3_client("s3", endpoint_url=os.getenv("S3_ENDPOINT_URL"))
    s3.download_file(bucket, key, filename)


def get_bucket_and_key(path: str) -> tuple[str, str]:
    """
    Get bucket and key from path.

    Parameters
    ----------
    path : str
        The source path to get the key from.

    Returns
    -------
    tuple[str, str]
        The bucket and key.
    """
    parsed = urlparse(path)
    return parsed.netloc, parsed.path


class MyEncoder(json.JSONEncoder):
    """
    Custom JSON encoder to handle numpy types.
    """

    def default(self, obj: Any) -> Any:
        """
        Convert numpy types to json.

        Parameters
        ----------
        obj : Any
            The object to convert.

        Returns
        -------
        Any
            The object converted to json.
        """
        if isinstance(obj, (int, str, float, list, dict)):
            return obj
        elif isinstance(obj, (np.integer, np.int64)):
            return int(obj)
        elif isinstance(obj, (np.floating, np.float64)):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return str(obj)


def dict_to_json(struct: dict) -> str:
    """
    Convert a dict to json.

    Parameters
    ----------
    struct : dict
        The dict to convert.

    Returns
    -------
    str
        The json string.
    """
    return json.dumps(struct, cls=MyEncoder)



from __future__ import annotations

from digitalhub_core.entities._base.spec.base import Spec, SpecParams
from digitalhub_core.entities.task.models import K8s


class TaskSpec(Spec):
    """Task specification."""

    def __init__(self, function: str) -> None:
        self.function = function


class TaskSpecK8s(TaskSpec):
    def __init__(
        self,
        function: str,
        **kwargs,
    ) -> None:
        super().__init__(function)
        self.node_selector = kwargs.get("node_selector")
        self.volumes = kwargs.get("volumes")
        self.resources = kwargs.get("resources")
        self.affinity = kwargs.get("affinity")
        self.tolerations = kwargs.get("tolerations")
        self.envs = kwargs.get("envs")
        self.secrets = kwargs.get("secrets")
        self.profile = kwargs.get("profile")


class TaskParams(SpecParams):
    """
    Base task model.
    """

    function: str
    """Function string."""


class TaskParamsK8s(TaskParams, K8s):
    """
    Kubernetes task model.
    """



from __future__ import annotations

from pydantic import BaseModel, Field
from typing_extensions import Literal


class Volume(BaseModel):
    """
    Volume model.
    """

    volume_type: Literal["persistent_volume_claim", "empty_dir"]
    """Volume type."""

    name: str
    """Volume name."""

    mount_path: str
    """Volume mount path inside the container."""

    spec: dict[str, str]
    """Volume spec."""


class NodeSelector(BaseModel):
    """
    NodeSelector model.
    """

    key: str
    """Node selector key."""

    value: str
    """Node selector value."""


class ResourceItem(BaseModel):
    """
    Resource item model.
    """

    requests: str = Field(default=None, regex=r"[\d]+|^([0-9])+([a-zA-Z])+$")
    """Resource requests."""

    limits: str = Field(default=None, regex=r"[\d]+|^([0-9])+([a-zA-Z])+$")
    """Resource limits."""


class Resource(BaseModel):
    """
    Resource model.
    """

    cpu: ResourceItem = None
    """CPU resource model."""

    mem: ResourceItem = None
    """Memory resource model."""

    gpu: ResourceItem = None
    """GPU resource model."""


class Env(BaseModel):
    """
    Env variable model.
    """

    name: str
    """Env variable name."""

    value: str
    """Env variable value."""


class Toleration(BaseModel):
    """
    Toleration model.
    """

    key: str = None
    """Toleration key."""

    operator: str = None
    """Toleration operator."""

    value: str = None
    """Toleration value."""

    effect: str = None
    """Toleration effect."""

    toleration_seconds: int = None
    """Toleration seconds."""


class V1NodeSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: list[str] = None


class V1NodeSelectorTerm(BaseModel):
    match_expressions: list[V1NodeSelectorRequirement] = None
    match_fields: list[V1NodeSelectorRequirement] = None


class V1NodeSelector(BaseModel):
    node_selector_terms: list[V1NodeSelectorTerm]


class V1PreferredSchedulingTerm(BaseModel):
    preference: V1NodeSelector
    weight: int


class V1LabelSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: list[str] = None


class V1LabelSelector(BaseModel):
    match_expressions: list[V1LabelSelectorRequirement] = None
    match_labels: dict[str, str] = None


class V1PodAffinityTerm(BaseModel):
    label_selector: V1LabelSelector = None
    match_label_keys: list[str] = None
    mismatch_label_keys: list[str] = None
    namespace_selector: V1LabelSelector = None
    namespaces: list[str] = None
    topology_key: str


class V1WeightedPodAffinityTerm(BaseModel):
    pod_affinity_term: V1PodAffinityTerm
    weight: int


class V1NodeAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1PreferredSchedulingTerm] = None
    required_during_scheduling_ignored_during_execution: V1NodeSelector = None


class V1PodAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1WeightedPodAffinityTerm] = None
    required_during_scheduling_ignored_during_execution: list[V1PodAffinityTerm] = None


class V1PodAntiAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: list[V1WeightedPodAffinityTerm] = None
    required_during_scheduling_ignored_during_execution: list[V1PodAffinityTerm] = None


class Affinity(BaseModel):
    """
    Affinity model.
    """

    node_affinity: V1NodeAffinity = None
    """Node affinity."""

    pod_affinity: V1PodAffinity = None
    """Pod affinity."""

    pod_anti_affinity: V1PodAntiAffinity = None
    """Pod anti affinity."""


class K8s(BaseModel):
    """
    Kubernetes resource model.
    """

    node_selector: list[NodeSelector] = None
    """Node selector."""

    volumes: list[Volume] = None
    """List of volumes."""

    resources: Resource = None
    """Resources restrictions."""

    affinity: Affinity = None
    """Affinity."""

    tolerations: list[Toleration] = None
    """Tolerations."""

    envs: list[Env] = None
    """Env variables."""

    secrets: list[str] = None
    """List of secret names."""

    profile: str = None
    """Profile template."""



from __future__ import annotations

import typing
from abc import abstractmethod
from typing import Any, Callable

from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub_core.runtimes.kind_registry import KindRegistry


class Runtime:
    """
    Base Runtime class.

    Runtimes are the entities responsible for the actual execution
    of a given run. They are highly specialized components which
    can translate the representation of a given execution as expressed
    in the run into an actual execution operation performed via
    libraries, code, external tools etc.
    """

    def __init__(self, kind_registry: KindRegistry, project: str) -> None:
        self.kind_registry = kind_registry
        self.project = project

    @abstractmethod
    def build(self, executable: dict, task: dict, run: dict) -> dict:
        """
        Build run spec.
        """

    @abstractmethod
    def run(self, run: dict) -> dict:
        """
        Execute run task.
        """

    @staticmethod
    @abstractmethod
    def _get_executable(action: str) -> Callable:
        """
        Get executable from action.
        """

    ##############################
    # Private methods
    ##############################

    def _validate_task(self, run: dict) -> str:
        """
        Check if task is allowed. This presumes that the
        runtime holds a list of allowed actions in the self.allowed_actions
        attribute.

        Parameters
        ----------
        run : dict
            Run object dictionary.

        Returns
        -------
        str
            Action to execute.
        """
        try:
            task_kind = run["spec"]["task"].split(":")[0]
        except (KeyError, IndexError):
            msg = "Malformed run spec."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

        try:
            return self.get_action_from_task_kind(task_kind)
        except EntityError:
            msg = f"Task {task_kind} not allowed."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    @staticmethod
    def _execute(func: Callable, *args, **kwargs) -> Any:
        """
        Execute function.

        Parameters
        ----------
        func : Callable
            Function to execute.
        *args
            Function arguments.
        **kwargs : dict
            Function keyword arguments.

        Returns
        -------
        Any
            Function result.
        """
        try:
            return func(*args, **kwargs)
        except Exception:
            msg = "Something got wrong during function execution."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    ##############################
    # Wrapper registry methods
    ##############################

    def get_executable_kind(self) -> str:
        """
        Get executable kind.

        Returns
        -------
        str
            Executable kind.
        """
        return self.kind_registry.get_executable_kind()

    def get_task_kind_from_action(self, action: str) -> str:
        """
        Get task kind from action.

        Parameters
        ----------
        action : str
            Task action.

        Returns
        -------
        str
            Task kind.
        """
        return self.kind_registry.get_task_kind_from_action(action)

    def get_action_from_task_kind(self, kind: str) -> str:
        """
        Get action from task.

        Parameters
        ----------
        kind : str
            Task kind.

        Returns
        -------
        str
            Action.
        """
        return self.kind_registry.get_action_from_task_kind(kind)

    def get_run_kind(self) -> str:
        """
        Get run kind.

        Returns
        -------
        str
            Run kind.
        """
        return self.kind_registry.get_run_kind()



from __future__ import annotations

from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.name import build_name
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.project.entity.core import ProjectCore as Project


def project_from_parameters(
    name: str,
    kind: str,
    description: str | None = None,
    labels: list[str] | None = None,
    local: bool = False,
    context: str | None = None,
    **kwargs,
) -> Project:
    """
    Create a new object.

    Parameters
    ----------
    name : str
        Object name.
    kind : str
        Kind the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    local : bool
        If True, use local backend, if False use DHCore backend. Default to False.
    context : str
        The context local folder of the project.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Project
        Object instance.
    """
    name = build_name(name)
    spec = build_spec(
        kind,
        context=context,
        **kwargs,
    )
    metadata = build_metadata(
        kind,
        project=name,
        name=name,
        description=description,
        labels=labels,
    )
    status = build_status(kind)
    return Project(
        name=name,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
        local=local,
    )


def project_from_dict(obj: dict) -> Project:
    """
    Create a new object from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create object from.

    Returns
    -------
    Project
        Object instance.
    """
    return Project.from_dict(obj)



from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj
from pydantic import BaseModel


class Spec(ModelObj):
    """
    A class representing the specification of an entity.
    Specification is a collection of information about an entity
    thought to be immutable by the user.
    """

    @classmethod
    def from_dict(cls, obj: dict) -> Spec:
        """
        Return entity specification object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity specification.

        Returns
        -------
        EntitySpec
            An entity specification object.
        """
        return cls(**obj)


class SpecParams(BaseModel, extra="ignore"):
    """
    A class representing the parameters of an entity.
    This base class is used to define the parameters of an entity
    specification and is used to validate the parameters passed
    to the constructor.
    """


class MaterialSpec(Spec):
    """
    Material specification class.
    """

    def __init__(self, path: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.path = path


class MaterialParams(SpecParams):
    """
    Material parameters class.
    """

    path: str
    """Target path to file(s)"""
